from __future__ import annotations

import os
import platform
from ctypes import (
    HRESULT,
    POINTER,
    Structure,
    c_void_p,
    c_wchar_p,
    windll,
)
from ctypes.wintypes import BOOL, FLOAT, UINT

from pyglet.font.dwrite.d2d1_types_lib import (
    D2D1_COLOR_F,
    D2D1_POINT_2F,
    D2D1_RECT_F,
    D2D1_SIZE_F,
    D2D1_SIZE_U,
    D2D_POINT_2F,
)
from pyglet.font.dwrite.dwrite_lib import (
    DWRITE_GLYPH_IMAGE_FORMATS,
    DWRITE_GLYPH_RUN,
    DWRITE_GLYPH_RUN_DESCRIPTION,
    DWRITE_MEASURING_MODE,
    IDWriteTextFormat,
    IDWriteTextLayout,
)
from pyglet.image.codecs.wincodec_lib import IWICBitmap, IWICBitmapSource
from pyglet.libs.win32 import c_void, com
from pyglet.libs.win32.types import BYTE, UINT32, UINT64

# --- Direct2D
try:
    d2d1 = "d2d1"

    # System32 and SysWOW64 folders are opposite perception in Windows x64.
    # System32 = x64 dll's | SysWOW64 = x86 dlls
    # By default ctypes only seems to look in system32 regardless of Python architecture, which has x64 dlls.
    if platform.architecture()[0] == "32bit" and platform.machine().endswith("64"):  # Machine is x64, Python is x86.
        d2d1 = os.path.join(os.environ["WINDIR"], "SysWOW64", "d2d1.dll")

    d2d_lib = windll.LoadLibrary(d2d1)
except OSError:
    # Doesn't exist? Should stop import of library.
    msg = "d2d1 Not Found"
    raise ImportError(msg)  # noqa: B904

D2D1_FACTORY_TYPE = UINT
D2D1_FACTORY_TYPE_SINGLE_THREADED = 0
D2D1_FACTORY_TYPE_MULTI_THREADED = 1

D2D1_COLOR_BITMAP_GLYPH_SNAP_OPTION = UINT
D2D1_COLOR_BITMAP_GLYPH_SNAP_OPTION_DEFAULT = 0
D2D1_COLOR_BITMAP_GLYPH_SNAP_OPTION_DISABLE = 1
D2D1_COLOR_BITMAP_GLYPH_SNAP_OPTION_FORCE_DWORD = 0xffffffff


D2D1_TEXT_ANTIALIAS_MODE = UINT
D2D1_TEXT_ANTIALIAS_MODE_DEFAULT = 0
D2D1_TEXT_ANTIALIAS_MODE_CLEARTYPE = 1
D2D1_TEXT_ANTIALIAS_MODE_GRAYSCALE = 2
D2D1_TEXT_ANTIALIAS_MODE_ALIASED = 3

D2D1_RENDER_TARGET_TYPE = UINT
D2D1_RENDER_TARGET_TYPE_DEFAULT = 0
D2D1_RENDER_TARGET_TYPE_SOFTWARE = 1
D2D1_RENDER_TARGET_TYPE_HARDWARE = 2

D2D1_FEATURE_LEVEL = UINT
D2D1_FEATURE_LEVEL_DEFAULT = 0

D2D1_RENDER_TARGET_USAGE = UINT
D2D1_RENDER_TARGET_USAGE_NONE = 0
D2D1_RENDER_TARGET_USAGE_FORCE_BITMAP_REMOTING = 1
D2D1_RENDER_TARGET_USAGE_GDI_COMPATIBLE = 2

DXGI_FORMAT = UINT
DXGI_FORMAT_UNKNOWN = 0

D2D1_ALPHA_MODE = UINT
D2D1_ALPHA_MODE_UNKNOWN = 0
D2D1_ALPHA_MODE_PREMULTIPLIED = 1
D2D1_ALPHA_MODE_STRAIGHT = 2
D2D1_ALPHA_MODE_IGNORE = 3

D2D1_DRAW_TEXT_OPTIONS = UINT
D2D1_DRAW_TEXT_OPTIONS_NO_SNAP = 0x00000001
D2D1_DRAW_TEXT_OPTIONS_CLIP = 0x00000002
D2D1_DRAW_TEXT_OPTIONS_ENABLE_COLOR_FONT = 0x00000004
D2D1_DRAW_TEXT_OPTIONS_DISABLE_COLOR_BITMAP_SNAPPING = 0x00000008
D2D1_DRAW_TEXT_OPTIONS_NONE = 0x00000000
D2D1_DRAW_TEXT_OPTIONS_FORCE_DWORD = 0xffffffff

DXGI_FORMAT_B8G8R8A8_UNORM = 87

D2D1_BITMAP_INTERPOLATION_MODE = UINT
D2D1_BITMAP_INTERPOLATION_MODE_NEAREST_NEIGHBOR = 0
D2D1_BITMAP_INTERPOLATION_MODE_LINEAR = 1
D2D1_BITMAP_INTERPOLATION_MODE_FORCE_DWORD = 0xffffffff

D2D1_BITMAP_OPTIONS = UINT
D2D1_BITMAP_OPTIONS_NONE = 0x00000000
D2D1_BITMAP_OPTIONS_TARGET = 0x00000001
D2D1_BITMAP_OPTIONS_CANNOT_DRAW = 0x00000002
D2D1_BITMAP_OPTIONS_CPU_READ = 0x00000004
D2D1_BITMAP_OPTIONS_GDI_COMPATIBLE = 0x00000008


D2D1CreateFactory = d2d_lib.D2D1CreateFactory
D2D1CreateFactory.restype = HRESULT
D2D1CreateFactory.argtypes = [D2D1_FACTORY_TYPE, com.REFIID, c_void_p, c_void_p]




class D2D1_PIXEL_FORMAT(Structure):
    _fields_ = (
        ("format", DXGI_FORMAT),
        ("alphaMode", D2D1_ALPHA_MODE),
    )


class D2D1_RENDER_TARGET_PROPERTIES(Structure):
    _fields_ = (
        ("type", D2D1_RENDER_TARGET_TYPE),
        ("pixelFormat", D2D1_PIXEL_FORMAT),
        ("dpiX", FLOAT),
        ("dpiY", FLOAT),
        ("usage", D2D1_RENDER_TARGET_USAGE),
        ("minLevel", D2D1_FEATURE_LEVEL),
    )

class D2D1_BITMAP_PROPERTIES(Structure):
    _fields_ = (
        ("pixelFormat", D2D1_PIXEL_FORMAT),
        ("dpiX", FLOAT),
        ("dpiY", FLOAT),
    )

ID2D1ColorContext = c_void_p
class D2D1_BITMAP_PROPERTIES1(Structure):
    _fields_ = (
        ("pixelFormat", D2D1_PIXEL_FORMAT),
        ("dpiX", FLOAT),
        ("dpiY", FLOAT),
        ("bitmapOptions", D2D1_BITMAP_OPTIONS),
        ("colorContext", ID2D1ColorContext),
    )

class ID2D1Resource(com.pIUnknown):
    _methods_ = [
        ("GetFactory",
         com.STDMETHOD()),
    ]


class ID2D1Brush(ID2D1Resource):
    _methods_ = [
        ("SetOpacity",
         com.STDMETHOD()),
        ("SetTransform",
         com.STDMETHOD()),
        ("GetOpacity",
         com.STDMETHOD()),
        ("GetTransform",
         com.STDMETHOD()),
    ]


class ID2D1SolidColorBrush(ID2D1Brush):
    _methods_ = [
        ("SetColor",
         com.METHOD(None, POINTER(D2D1_COLOR_F))),
        ("GetColor",
         com.STDMETHOD()),
    ]

D2D1_DEVICE_CONTEXT_OPTIONS = UINT
D2D1_DEVICE_CONTEXT_OPTIONS_NONE = 0
D2D1_DEVICE_CONTEXT_OPTIONS_ENABLE_MULTITHREADED_OPTIMIZATIONS = 1
D2D1_DEVICE_CONTEXT_OPTIONS_FORCE_DWORD = 0xffffffff

D2D1_TAG = UINT64

D2D1_MAP_OPTIONS = UINT
D2D1_MAP_OPTIONS_NONE = 0
D2D1_MAP_OPTIONS_READ = 1
D2D1_MAP_OPTIONS_WRITE = 2
D2D1_MAP_OPTIONS_DISCARD = 4

class D2D1_MAPPED_RECT(Structure):
    _fields_ = [
        ("pitch", UINT32),
        ("bits", POINTER(BYTE)),
    ]

class ID2D1Image(ID2D1Resource):
    _methods_ = []

class ID2D1Bitmap(ID2D1Image):
    _methods_ = [
        ("GetSize",
         com.METHOD(D2D1_SIZE_F)),
        ("GetPixelSize",
         com.METHOD(D2D1_SIZE_U)),
        ("GetPixelFormat",
         com.METHOD(D2D1_PIXEL_FORMAT)),
        ("GetDpi",
         com.METHOD(c_void, POINTER(FLOAT), POINTER(FLOAT))),
        ("CopyFromBitmap",
         com.STDMETHOD()),
        ("CopyFromRenderTarget",
         com.STDMETHOD()),
        ("CopyFromMemory",
         com.STDMETHOD()),
    ]

class ID2D1Bitmap1(ID2D1Bitmap):
    _methods_ = [
        ("GetColorContext",
         com.STDMETHOD()),
        ("GetOptions",
         com.METHOD(D2D1_BITMAP_OPTIONS)),
        ("GetSurface",
         com.STDMETHOD()),
        ("Map",
         com.STDMETHOD(D2D1_MAP_OPTIONS, POINTER(D2D1_MAPPED_RECT))),
        ("Unmap",
         com.STDMETHOD()),
    ]


class IDXGISurface(com.pIUnknown):
    ...

class ID2D1RenderTarget(ID2D1Resource):
    _methods_ = [
        ("CreateBitmap",
         com.STDMETHOD(D2D1_SIZE_U, c_void_p, UINT32, POINTER(D2D1_BITMAP_PROPERTIES), POINTER(ID2D1Bitmap))),
        ("CreateBitmapFromWicBitmap",
         com.STDMETHOD(IWICBitmapSource, POINTER(D2D1_BITMAP_PROPERTIES), POINTER(ID2D1Bitmap))),
        ("CreateSharedBitmap",
         com.STDMETHOD()),
        ("CreateBitmapBrush",
         com.STDMETHOD()),
        ("CreateSolidColorBrush",
         com.STDMETHOD(POINTER(D2D1_COLOR_F), c_void_p, POINTER(ID2D1SolidColorBrush))),
        ("CreateGradientStopCollection",
         com.STDMETHOD()),
        ("CreateLinearGradientBrush",
         com.STDMETHOD()),
        ("CreateRadialGradientBrush",
         com.STDMETHOD()),
        ("CreateCompatibleRenderTarget",
         com.STDMETHOD()),
        ("CreateLayer",
         com.STDMETHOD()),
        ("CreateMesh",
         com.STDMETHOD()),
        ("DrawLine",
         com.STDMETHOD()),
        ("DrawRectangle",
         com.STDMETHOD()),
        ("FillRectangle",
         com.STDMETHOD()),
        ("DrawRoundedRectangle",
         com.STDMETHOD()),
        ("FillRoundedRectangle",
         com.STDMETHOD()),
        ("DrawEllipse",
         com.STDMETHOD()),
        ("FillEllipse",
         com.STDMETHOD()),
        ("DrawGeometry",
         com.STDMETHOD()),
        ("FillGeometry",
         com.STDMETHOD()),
        ("FillMesh",
         com.STDMETHOD()),
        ("FillOpacityMask",
         com.METHOD(None, ID2D1Bitmap, ID2D1Brush, UINT, POINTER(D2D1_RECT_F), POINTER(D2D1_RECT_F))),
        ("DrawBitmap",
         com.STDMETHOD(ID2D1Bitmap, POINTER(D2D1_RECT_F), FLOAT, D2D1_BITMAP_INTERPOLATION_MODE, POINTER(D2D1_RECT_F))),
        ("DrawText",
         com.STDMETHOD()),
        ("DrawTextLayout",
         com.METHOD(c_void, D2D_POINT_2F, IDWriteTextLayout, ID2D1Brush, UINT32)),
        ("DrawGlyphRun",
         com.METHOD(c_void, D2D_POINT_2F, POINTER(DWRITE_GLYPH_RUN), ID2D1Brush, UINT32)),
        ("SetTransform",
         com.METHOD(c_void)),
        ("GetTransform",
         com.STDMETHOD()),
        ("SetAntialiasMode",
         com.METHOD(c_void, D2D1_TEXT_ANTIALIAS_MODE)),
        ("GetAntialiasMode",
         com.STDMETHOD()),
        ("SetTextAntialiasMode",
         com.METHOD(c_void, D2D1_TEXT_ANTIALIAS_MODE)),
        ("GetTextAntialiasMode",
         com.STDMETHOD()),
        ("SetTextRenderingParams",
         com.STDMETHOD()),
        ("GetTextRenderingParams",
         com.STDMETHOD()),
        ("SetTags",
         com.STDMETHOD()),
        ("GetTags",
         com.STDMETHOD()),
        ("PushLayer",
         com.STDMETHOD()),
        ("PopLayer",
         com.STDMETHOD()),
        ("Flush",
         com.STDMETHOD(POINTER(D2D1_TAG), POINTER(D2D1_TAG))),
        ("SaveDrawingState",
         com.STDMETHOD()),
        ("RestoreDrawingState",
         com.STDMETHOD()),
        ("PushAxisAlignedClip",
         com.STDMETHOD()),
        ("PopAxisAlignedClip",
         com.STDMETHOD()),
        ("Clear",
         com.METHOD(c_void, POINTER(D2D1_COLOR_F))),
        ("BeginDraw",
         com.METHOD(c_void)),
        ("EndDraw",
         com.STDMETHOD(POINTER(D2D1_TAG), POINTER(D2D1_TAG))),
        ("GetPixelFormat",
         com.STDMETHOD()),
        ("SetDpi",
         com.STDMETHOD(FLOAT, FLOAT)),
        ("GetDpi",
         com.STDMETHOD()),
        ("GetSize",
         com.STDMETHOD()),
        ("GetPixelSize",
         com.STDMETHOD()),
        ("GetMaximumBitmapSize",
         com.STDMETHOD()),
        ("IsSupported",
         com.METHOD(BOOL, POINTER(D2D1_RENDER_TARGET_PROPERTIES))),
    ]


class ID2D1DeviceContext(ID2D1RenderTarget):
    _methods_ = [
        ("CreateBitmap_Context",
         com.STDMETHOD(D2D1_SIZE_U, c_void_p, UINT32, POINTER(D2D1_BITMAP_PROPERTIES1), POINTER(ID2D1Bitmap1))),
        ("CreateBitmapFromWicBitmap_Context",
         com.STDMETHOD(IWICBitmapSource, POINTER(D2D1_BITMAP_PROPERTIES1), POINTER(ID2D1Bitmap1))),
        ("CreateColorContext",
         com.STDMETHOD()),
        ("CreateColorContextFromFilename",
         com.STDMETHOD()),
        ("CreateColorContextFromWicColorContext",
         com.STDMETHOD()),
        ("CreateBitmapFromDxgiSurface",
         com.STDMETHOD(IDXGISurface, POINTER(D2D1_BITMAP_PROPERTIES1), POINTER(ID2D1Bitmap1))),
        ("CreateEffect",
         com.STDMETHOD()),
        ("CreateGradientStopCollection_Context",
         com.STDMETHOD()),
        ("CreateImageBrush",
         com.STDMETHOD()),
        ("CreateBitmapBrush_Context",
         com.STDMETHOD()),
        ("CreateCommandList",
         com.STDMETHOD()),
        ("IsDxgiFormatSupported",
         com.STDMETHOD()),
        ("IsBufferPrecisionSupported",
         com.STDMETHOD()),
        ("GetImageLocalBounds",
         com.STDMETHOD()),
        ("GetImageWorldBounds",
         com.STDMETHOD()),
        ("GetGlyphRunWorldBounds",
         com.STDMETHOD()),
        ("GetDevice",
         com.STDMETHOD()),
        ("SetTarget",
         com.STDMETHOD(c_void_p)),
        ("GetTarget",
         com.METHOD(c_void, POINTER(c_void_p))),
        ("SetRenderingControls",
         com.STDMETHOD()),
        ("GetRenderingControls",
         com.STDMETHOD()),
        ("SetPrimitiveBlend",
         com.STDMETHOD()),
        ("GetPrimitiveBlend",
         com.STDMETHOD()),
        ("SetUnitMode",
         com.STDMETHOD()),
        ("GetUnitMode",
         com.STDMETHOD()),
        ("DrawGlyphRun_Context",
         com.METHOD(c_void, D2D_POINT_2F, POINTER(DWRITE_GLYPH_RUN), POINTER(DWRITE_GLYPH_RUN_DESCRIPTION), ID2D1Brush, UINT32)),
        ("DrawImage",
         com.STDMETHOD()),
        ("DrawGdiMetafile",
         com.STDMETHOD()),
        ("DrawBitmap_Context",
         com.STDMETHOD()),
        ("PushLayer_Context",
         com.STDMETHOD()),
        ("InvalidateEffectInputRectangle",
         com.STDMETHOD()),
        ("GetEffectInvalidRectangleCount",
         com.STDMETHOD()),
        ("GetEffectInvalidRectangles",
         com.STDMETHOD()),
        ("GetEffectRequiredInputRectangles",
         com.STDMETHOD()),
        ("FillOpacityMask_Context",
         com.METHOD(None, ID2D1Bitmap, ID2D1Brush, POINTER(D2D1_RECT_F), POINTER(D2D1_RECT_F))),
    ]

class ID2D1DeviceContext1(ID2D1DeviceContext):
    _methods_ = [
        ("CreateFilledGeometryRealization",
         com.STDMETHOD()),
        ("CreateStrokedGeometryRealization",
         com.STDMETHOD()),
        ("DrawGeometryRealization",
         com.STDMETHOD()),
    ]

class ID2D1DeviceContext2(ID2D1DeviceContext1):
    _methods_ = [
        ("CreateInk",
         com.STDMETHOD()),
        ("CreateInkStyle",
         com.STDMETHOD()),
        ("CreateGradientMesh",
         com.STDMETHOD()),
        ("CreateImageSourceFromWic",
         com.STDMETHOD()),
        ("CreateLookupTable3D",
         com.STDMETHOD()),
        ("CreateImageSourceFromDxgi",
         com.STDMETHOD()),
        ("GetGradientMeshWorldBounds",
         com.STDMETHOD()),
        ("DrawInk",
         com.STDMETHOD()),
        ("DrawGradientMesh",
         com.STDMETHOD()),
        ("DrawGdiMetafile2",
         com.STDMETHOD()),
        ("CreateTransformedImageSource",
         com.STDMETHOD()),
    ]

class ID2D1DeviceContext3(ID2D1DeviceContext2):
    _methods_ = [
        ("CreateSpriteBatch",
         com.STDMETHOD()),
        ("DrawSpriteBatch",
         com.STDMETHOD()),
    ]

ID2D1SvgGlyphStyle = c_void_p
class ID2D1DeviceContext4(ID2D1DeviceContext3):
    _methods_ = [
        ("CreateSvgGlyphStyle",
         com.STDMETHOD()),
        ("DrawText4",
         com.STDMETHOD(c_wchar_p, UINT32, IDWriteTextFormat, D2D1_RECT_F, ID2D1Brush, ID2D1SvgGlyphStyle, UINT32, D2D1_DRAW_TEXT_OPTIONS, DWRITE_MEASURING_MODE)),
        ("DrawTextLayout4",
         com.STDMETHOD(D2D1_POINT_2F, IDWriteTextLayout, ID2D1Brush, ID2D1SvgGlyphStyle, UINT32, D2D1_DRAW_TEXT_OPTIONS)),
        ("DrawColorBitmapGlyphRun",
         com.METHOD(c_void, DWRITE_GLYPH_IMAGE_FORMATS, D2D1_POINT_2F, POINTER(DWRITE_GLYPH_RUN), DWRITE_MEASURING_MODE, D2D1_COLOR_BITMAP_GLYPH_SNAP_OPTION)),
        ("DrawSvgGlyphRun",
         com.METHOD(c_void, D2D_POINT_2F, POINTER(DWRITE_GLYPH_RUN), ID2D1Brush, ID2D1SvgGlyphStyle, UINT32, DWRITE_MEASURING_MODE)),
        ("GetColorBitmapGlyphImage",
         com.STDMETHOD()),
        ("GetSvgGlyphImage",
         com.STDMETHOD()),
    ]





class ID2D1Device(ID2D1Resource):
    _methods_ = [
        ("CreateDeviceContext", com.STDMETHOD(D2D1_DEVICE_CONTEXT_OPTIONS, POINTER(ID2D1DeviceContext))),
        ("CreatePrintControl", com.STDMETHOD()),
        ("SetMaximumTextureMemory", com.STDMETHOD()),
        ("GetMaximumTextureMemory", com.STDMETHOD()),
        ("ClearResources", com.STDMETHOD()),
    ]

class ID2D1Device1(ID2D1Device):
    _methods_ = [
        ("GetRenderingPriority", com.STDMETHOD()),  # D2D1_RENDERING_PRIORITY GetRenderingPriority()
        ("SetRenderingPriority", com.STDMETHOD()),
        ("CreateDeviceContext1", com.STDMETHOD(D2D1_DEVICE_CONTEXT_OPTIONS, POINTER(ID2D1DeviceContext1))),
    ]

class ID2D1Device2(ID2D1Device1):
    _methods_ = [
        ("CreateDeviceContext2", com.STDMETHOD()),
        ("FlushDeviceContexts", com.STDMETHOD()),
        ("GetDxgiDevice", com.STDMETHOD()),
    ]

class ID2D1Device3(ID2D1Device2):
    _methods_ = [
        ("CreateDeviceContext3", com.STDMETHOD()),
    ]

class ID2D1Device4(ID2D1Device3):
    _methods_ = [
        ("CreateDeviceContext4", com.STDMETHOD(D2D1_DEVICE_CONTEXT_OPTIONS, POINTER(ID2D1DeviceContext4))),
        ("SetMaximumColorGlyphCacheMemory", com.METHOD(c_void, UINT64)),
        ("GetMaximumColorGlyphCacheMemory", com.METHOD(UINT64)),
    ]



class ID2D1Factory(com.pIUnknown):
    _methods_ = [
        ("ReloadSystemMetrics",
         com.STDMETHOD()),
        ("GetDesktopDpi",
         com.STDMETHOD()),
        ("CreateRectangleGeometry",
         com.STDMETHOD()),
        ("CreateRoundedRectangleGeometry",
         com.STDMETHOD()),
        ("CreateEllipseGeometry",
         com.STDMETHOD()),
        ("CreateGeometryGroup",
         com.STDMETHOD()),
        ("CreateTransformedGeometry",
         com.STDMETHOD()),
        ("CreatePathGeometry",
         com.STDMETHOD()),
        ("CreateStrokeStyle",
         com.STDMETHOD()),
        ("CreateDrawingStateBlock",
         com.STDMETHOD()),
        ("CreateWicBitmapRenderTarget",
         com.STDMETHOD(IWICBitmap, POINTER(D2D1_RENDER_TARGET_PROPERTIES), POINTER(ID2D1RenderTarget))),
        ("CreateHwndRenderTarget",
         com.STDMETHOD()),
        ("CreateDxgiSurfaceRenderTarget",
         com.STDMETHOD()),
        ("CreateDCRenderTarget",
         com.STDMETHOD()),
    ]


class ID2D1Factory1(ID2D1Factory):
    _methods_ = [
        ("CreateDevice", com.STDMETHOD(c_void_p, POINTER(ID2D1Device))),
        ("CreateStrokeStyle1", com.STDMETHOD()),
        ("CreatePathGeometry1", com.STDMETHOD()),
        ("CreateDrawingStateBlock1", com.STDMETHOD()),
        ("CreateGdiMetafile", com.STDMETHOD()),
        ("RegisterEffectFromStream", com.STDMETHOD()),
        ("RegisterEffectFromString", com.STDMETHOD()),
        ("UnregisterEffect", com.STDMETHOD()),
        ("GetRegisteredEffects", com.STDMETHOD()),
        ("GetEffectProperties", com.STDMETHOD()),
    ]

class ID2D1Factory2(ID2D1Factory1):
    _methods_ = [
        ("CreateDevice2", com.STDMETHOD()),
    ]

class ID2D1Factory3(ID2D1Factory2):
    _methods_ = [
        ("CreateDevice3", com.STDMETHOD()),
    ]


class ID2D1Factory4(ID2D1Factory3):
    _methods_ = [
        ("CreateDevice4", com.STDMETHOD()),
    ]

class ID2D1Factory5(ID2D1Factory4):
    _methods_ = [
        ("CreateDevice5",
         com.STDMETHOD(c_void_p, POINTER(ID2D1Device4))),
    ]


IID_ID2D1Factory = com.GUID(0x06152247, 0x6f50, 0x465a, 0x92, 0x45, 0x11, 0x8b, 0xfd, 0x3b, 0x60, 0x07)
IID_ID2D1Factory1 = com.GUID(0xbb12d362,0xdaee,0x4b9a,0xaa,0x1d,0x14,0xba,0x40,0x1c,0xfa,0x1f)
#IID_ID2D1Factory2 = com.GUID.from_string("94f81a73-9212-4376-9c58-b16a3a0d3992")
IID_ID2D1Factory4 = com.GUID.from_string("bd4ec2d2-0662-4bee-ba8e-6f29f032e096")
IID_ID2D1Factory5 = com.GUID.from_string("c4349994-838e-4b0f-8cab-44997d9eeacc")

# Default render parameters for D2D.
default_target_properties = D2D1_RENDER_TARGET_PROPERTIES(
    type=D2D1_RENDER_TARGET_TYPE_DEFAULT,
    pixelFormat=D2D1_PIXEL_FORMAT(
        format=DXGI_FORMAT_UNKNOWN,
        alphaMode=D2D1_ALPHA_MODE_UNKNOWN,
    ),
    dpiX=0.0,
    dpiY=0.0,
    usage=D2D1_RENDER_TARGET_USAGE_NONE,
    min_level=D2D1_FEATURE_LEVEL_DEFAULT,
)

default_bitmap_properties = D2D1_BITMAP_PROPERTIES1(
    pixelFormat=D2D1_PIXEL_FORMAT(
        format=DXGI_FORMAT_B8G8R8A8_UNORM,
        alphaMode=D2D1_ALPHA_MODE_PREMULTIPLIED,
    ),
    dpiX=0.0,
    dpiY=0.0,
    bitmapOptions=D2D1_BITMAP_OPTIONS_TARGET | D2D1_BITMAP_OPTIONS_CANNOT_DRAW,
)

default_bitmap_properties0 = D2D1_BITMAP_PROPERTIES(
    pixelFormat=D2D1_PIXEL_FORMAT(
        format=DXGI_FORMAT_B8G8R8A8_UNORM,
        alphaMode=D2D1_ALPHA_MODE_PREMULTIPLIED,
    ),
    dpiX=0.0,
    dpiY=0.0,
)

IID_ID2D1DeviceContext = com.GUID(0xe8f7fe7a, 0x191c, 0x466d, 0xad, 0x95, 0x97, 0x56, 0x78, 0xbd, 0xa9, 0x98)
IID_ID2D1DeviceContext4 = com.GUID.from_string("8c427831-3d90-4476-b647-c4fae349e4db")



from collections import deque
from typing import TYPE_CHECKING, List, Optional, Tuple
import weakref

from pyglet.media.drivers.base import AbstractAudioDriver, AbstractAudioPlayer, MediaEvent
from pyglet.media.drivers.listener import AbstractListener
from pyglet.media.drivers.openal import interface
from pyglet.media.player_worker_thread import PlayerWorkerThread
from pyglet.util import debug_print

if TYPE_CHECKING:
    from pyglet.media import Source, Player


_debug = debug_print('debug_media')


class OpenALDriver(AbstractAudioDriver):
    def __init__(self, device_name: Optional[str] = None) -> None:
        super().__init__()

        self.device = interface.OpenALDevice(device_name)
        self.context = self.device.create_context()
        self.context.make_current()

        self._listener = OpenALListener(self)

        self.worker = PlayerWorkerThread()
        self.worker.start()

    def create_audio_player(self, source: 'Source', player: 'Player') -> 'OpenALAudioPlayer':
        assert self.device is not None, 'Device was closed'
        return OpenALAudioPlayer(self, source, player)

    def delete(self) -> None:
        if self.context is None:
            assert _debug('Duplicate OpenALDriver.delete(), ignoring')
            return

        assert _debug("Delete OpenALDriver")
        self.worker.stop()

        # A device may only be closed if no more contexts and no more buffers exist on it
        # A buffer may only be deleted if no source is using it anymore
        # A context may only be deleted if it is free of sources
        # Buffers need a context to report errors to when being deleted
        self.context.delete_sources()
        self.device.buffer_pool.delete()
        self.context.delete()
        self.device.close()
        self.context = None

    def have_version(self, major: int, minor: int) -> bool:
        return (major, minor) <= self.get_version()

    def get_version(self) -> Tuple[int, int]:
        assert self.device is not None, "Device was closed"
        return self.device.get_version()

    def get_extensions(self) -> List[str]:
        assert self.device is not None, "Device was closed"
        return self.device.get_extensions()

    def have_extension(self, extension: str) -> bool:
        return extension in self.get_extensions()

    def get_listener(self) -> 'OpenALListener':
        return self._listener


class OpenALListener(AbstractListener):
    def __init__(self, driver: 'OpenALDriver') -> None:
        self._driver = weakref.proxy(driver)
        self._al_listener = interface.OpenALListener()

    def _set_volume(self, volume: float) -> None:
        self._al_listener.gain = volume
        self._volume = volume

    def _set_position(self, position: Tuple[float, float, float]) -> None:
        self._al_listener.position = position
        self._position = position

    def _set_forward_orientation(self, orientation: Tuple[float, float, float]) -> None:
        self._al_listener.orientation = orientation + self._up_orientation
        self._forward_orientation = orientation

    def _set_up_orientation(self, orientation: Tuple[float, float, float]) -> None:
        self._al_listener.orientation = self._forward_orientation + orientation
        self._up_orientation = orientation


class OpenALAudioPlayer(AbstractAudioPlayer):
    def __init__(self, driver: 'OpenALDriver', source: 'Source', player: 'Player') -> None:
        super().__init__(source, player)
        self.driver = driver
        self.alsource = driver.context.create_source()

        # Cursor positions, like DSound and Pulse drivers, refer to a
        # hypothetical infinite-length buffer.  Cursor units are in bytes.

        # The following should be true at all times:
        # buffer <= play <= write; buffer >= 0; play >= 0; write >= 0

        # Start of the current (head) AL buffer
        self._buffer_cursor = 0

        # Estimated playback cursor position (last seen)
        self._play_cursor = 0

        # Cursor position of end of the last queued AL buffer.
        self._write_cursor = 0

        # Whether the source has been exhausted of all data.
        # Don't bother trying to refill then and brace for eos.
        self._pyglet_source_exhausted = False

        # Whether the OpenAL source has played to its end.
        # Prevent duplicate dispatches of on_eos events.
        self._has_underrun = False

        # Deque of the currently queued buffer's sizes
        self._queued_buffer_sizes = deque()

    def delete(self) -> None:
        if self.alsource is not None:
            self.driver.worker.remove(self)
            self.alsource.delete()
            self.alsource = None

    def play(self) -> None:
        assert _debug('OpenALAudioPlayer.play()')
        assert self.driver is not None
        assert self.alsource is not None

        if not self.alsource.is_playing:
            self.alsource.play()

        self.driver.worker.add(self)

    def stop(self) -> None:
        assert _debug('OpenALAudioPlayer.stop()')
        assert self.driver is not None
        assert self.alsource is not None

        self.driver.worker.remove(self)
        self.alsource.pause()

    def clear(self) -> None:
        assert _debug('OpenALAudioPlayer.clear()')
        assert self.driver is not None
        assert self.alsource is not None

        super().clear()
        self.alsource.stop()
        self.alsource.clear()

        self._buffer_cursor = 0
        self._play_cursor = 0
        self._write_cursor = 0
        self._pyglet_source_exhausted = False
        self._has_underrun = False
        self._queued_buffer_sizes.clear()

    def _check_processed_buffers(self) -> None:
        buffers_processed = self.alsource.unqueue_buffers()
        for _ in range(buffers_processed):
            # Buffers have been processed (and already been removed from the ALSource);
            # Adjust buffer cursor.
            self._buffer_cursor += self._queued_buffer_sizes.popleft()

    def _update_play_cursor(self) -> None:
        self._play_cursor = self._buffer_cursor + self.alsource.byte_offset

    def work(self) -> None:
        self._check_processed_buffers()
        self._update_play_cursor()
        self.dispatch_media_events(self._play_cursor)

        if self._pyglet_source_exhausted:
            if not self._has_underrun and not self.alsource.is_playing:
                self._has_underrun = True
                assert _debug('OpenALAudioPlayer: Dispatching eos')
                MediaEvent('on_eos').sync_dispatch_to_player(self.player)
            return

        refilled = self._maybe_refill()

        if refilled and not self.alsource.is_playing:
            # Very unlikely case where the refill was delayed by so much the
            # source underran and stopped. If it did, restart it.
            self.alsource.play()

    def _maybe_refill(self) -> bool:
        if self._pyglet_source_exhausted:
            return False

        remaining_bytes = self._write_cursor - self._play_cursor
        if remaining_bytes >= self._buffered_data_comfortable_limit:
            return False

        missing_bytes = self._buffered_data_ideal_size - remaining_bytes
        self._refill(self.source.audio_format.align_ceil(missing_bytes))
        return True

    def get_play_cursor(self) -> int:
        return self._play_cursor

    def _refill(self, refill_size) -> None:
        audio_data = self._get_and_compensate_audio_data(refill_size, self._play_cursor)

        if audio_data is None:
            self._pyglet_source_exhausted = True
            return

        # We got new audio data; first queue its events
        self.append_events(self._write_cursor, audio_data.events)

        # Get, fill and queue OpenAL buffer using the entire AudioData
        buf = self.alsource.get_buffer()
        buf.data(audio_data, self.source.audio_format)
        self.alsource.queue_buffer(buf)

        # Adjust the write cursor and memorize buffer length
        self._write_cursor += audio_data.length
        self._queued_buffer_sizes.append(audio_data.length)

    def prefill_audio(self) -> None:
        self._maybe_refill()

    def set_volume(self, volume: float) -> None:
        self.alsource.gain = volume

    def set_position(self, position: Tuple[float, float, float]) -> None:
        self.alsource.position = position

    def set_min_distance(self, min_distance: float) -> None:
        self.alsource.reference_distance = min_distance

    def set_max_distance(self, max_distance: float) -> None:
        self.alsource.max_distance = max_distance

    def set_pitch(self, pitch: float) -> None:
        self.alsource.pitch = pitch

    def set_cone_orientation(self, cone_orientation: Tuple[float, float, float]) -> None:
        self.alsource.direction = cone_orientation

    def set_cone_inner_angle(self, cone_inner_angle: float) -> None:
        self.alsource.cone_inner_angle = cone_inner_angle

    def set_cone_outer_angle(self, cone_outer_angle: float) -> None:
        self.alsource.cone_outer_angle = cone_outer_angle

    def set_cone_outer_gain(self, cone_outer_gain: float) -> None:
        self.alsource.cone_outer_gain = cone_outer_gain



"""Game Controller support.

This module provides an interface for Game Controller devices, which are a
subset of Joysticks. Game Controllers have consistent button and axis mapping,
which resembles common dual-stick home video game console controllers.
Devices that are of this design can be automatically mapped to the "virtual"
Game Controller layout, providing a consistent abstraction for a large number
of different devices, with no tedious button and axis mapping for each one.
To achieve this, an internal mapping database contains lists of device ids
and their corresponding button and axis mappings. The mapping database is in
the same format as originated by the `SDL` library, which has become a
semi-standard and is in common use. Most popular controllers are included in
the built-in database, and additional mappings can be added at runtime.

Some Joysticks, such as Flight Sticks, etc., do not necessarily fit into the
layout (and limitations) of GameControllers. For those such devices, it is
recommended to use the Joystick interface instead.

To query which GameControllers are available, call :py:func:`get_controllers`.

.. versionadded:: 2.0
"""
from __future__ import annotations

import os as _os
import sys as _sys
import warnings as _warnings

from .base import Sign
from .controller_db import mapping_list


if _env_config := _os.environ.get('SDL_GAMECONTROLLERCONFIG'):
    # insert at the front of the list
    mapping_list.insert(0, _env_config)


def _swap_le16(value: int) -> int:
    """Ensure 16bit value is in Big Endian format"""
    if _sys.byteorder == "little":
        return ((value << 8) | (value >> 8)) & 0xFFFF
    return value


def create_guid(bus: int, vendor: int, product: int, version: int, name: str, signature: int, data: int) -> str:
    # byte size      16           16            16            16         str             8          8
    """Create an SDL2 style GUID string from a device's identifiers."""
    bus = _swap_le16(bus)
    vendor = _swap_le16(vendor)
    product = _swap_le16(product)
    version = _swap_le16(version)

    return f"{bus:04x}0000{vendor:04x}0000{product:04x}0000{version:04x}0000"


class Relation:
    __slots__ = 'control_type', 'index', 'sign'

    def __init__(self, control_type: str, index: int, sign: Sign = Sign.DEFAULT):
        self.control_type = control_type
        self.index = index
        self.sign = sign

    def __repr__(self):
        return f"Relation(type={self.control_type}, index={self.index}, sign={self.sign})"


def _parse_mapping(mapping_string: str) -> dict[str, str | Relation] | None:
    """Parse a SDL2 style GameController mapping string.

    Args:
        mapping_string: A string containing an SDL style controller mapping.
    """

    valid_keys = ['guide', 'back', 'start', 'a', 'b', 'x', 'y',
                  'leftshoulder', 'leftstick', 'rightshoulder', 'rightstick',
                  'dpup', 'dpdown', 'dpleft', 'dpright',
                  'lefttrigger', 'righttrigger', 'leftx', 'lefty', 'rightx', 'righty']

    try:
        guid, name, *split_mapping = mapping_string.strip().split(",")
        relations = dict(guid=guid, name=name)
    except ValueError:
        return None

    for item in split_mapping:
        # looking for items like: a:b0, b:b1, etc.
        if ':' not in item:
            continue

        key, relation_string, *etc = item.split(':')

        if key not in valid_keys:
            continue

        # Look for specific flags to signify axis sign:
        if "+" in relation_string:
            relation_string = relation_string.strip('+')
            sign = Sign.POSITIVE
        elif "-" in relation_string:
            relation_string = relation_string.strip('-')
            sign = Sign.NEGATIVE
        elif "~" in relation_string:
            relation_string = relation_string.strip('~')
            sign = Sign.INVERTED
        else:
            sign = Sign.DEFAULT

        # All relations will be one of (Button, Axis, or Hat).
        if relation_string.startswith("b"):  # Button
            relations[key] = Relation("button", int(relation_string[1:]), sign)
        elif relation_string.startswith("a"):  # Axis
            relations[key] = Relation("axis", int(relation_string[1:]), sign)
        elif relation_string.startswith("h0"):  # Hat
            relations[key] = Relation("hat0", int(relation_string.split(".")[1]), sign)

    return relations


def get_mapping(guid: str) -> dict[str, str | Relation] | None:
    """Return a mapping for the passed device GUID.

    Args:
        guid: A device GUID; see :py:meth:`~pyglet.input.Device.get_guid`
    """
    for mapping in mapping_list:
        if not mapping.startswith(guid):
            continue

        try:
            return _parse_mapping(mapping)
        except ValueError:
            _warnings.warn(f"Unable to parse Controller mapping: {mapping}")
            continue


def add_mappings_from_file(filename: str) -> None:
    """Add one or more mappings from a file.

    Given a file path, open and parse the file for mappings.

    Args:
        filename: A file path.
    """
    with open(filename) as f:
        add_mappings_from_string(f.read())


def add_mappings_from_string(string: str) -> None:
    """Add one or more mappings from a string.

        Args:
            string: A string containing one or more mappings,
        """
    for line in string.splitlines():
        line = line.strip()
        if line.startswith('#'):
            continue
        mapping_list.insert(0, line)



import pyglet
from game import load, player

# Set up a window
game_window = pyglet.window.Window(800, 600)

main_batch = pyglet.graphics.Batch()

# Set up the two top labels
score_label = pyglet.text.Label(text="Score: 0", x=10, y=575, batch=main_batch)
level_label = pyglet.text.Label(text="Version 2: Basic Motion",
                                x=400, y=575, anchor_x='center', batch=main_batch)

# Initialize the player sprite
player_ship = player.Player(x=400, y=300, batch=main_batch)

# Make three sprites to represent remaining lives
player_lives = load.player_lives(2, main_batch)

# Make three asteroids so we have something to shoot at
asteroids = load.asteroids(3, player_ship.position, main_batch)

# Store all objects that update each frame in a list
game_objects = [player_ship] + asteroids

# Tell the main window that the player object responds to events
game_window.push_handlers(player_ship)


@game_window.event
def on_draw():
    game_window.clear()
    main_batch.draw()


def update(dt):
    for obj in game_objects:
        obj.update(dt)


if __name__ == "__main__":
    # Update the game 120 times per second
    pyglet.clock.schedule_interval(update, 1 / 120.0)

    # Tell pyglet to do its thing
    pyglet.app.run()



#!/usr/bin/env python
"""Convert an image to another file format supported by pyglet.

Usage::
    python image_convert.py <src-file> <dest-file>

"""


import sys

import pyglet


def convert(src, dest):
    if '.dds' in src.lower():
        # Compressed textures need to be uploaded to the video card before
        # they can be saved.
        texture = pyglet.image.load(src).get_texture()
        texture.save(dest)
    else:
        # Otherwise just save the loaded image in the new format.
        image = pyglet.image.load(src)
        image.save(dest)


if __name__ == '__main__':
    if len(sys.argv) != 3:
        print(__doc__)
        sys.exit(1)

    src = sys.argv[1]
    dest = sys.argv[2]
    convert(src, dest)



import pyglet

# Import the new codec
import aseprite


window = pyglet.window.Window()
batch = pyglet.graphics.Batch()

# Add image decoders from the loaded codec module.
# This adds support to the image and resource modules.
pyglet.image.codecs.add_decoders(aseprite)

# You can now load a animation of *.ase type:
image = pyglet.image.load_animation("running.ase")
sprite = pyglet.sprite.Sprite(img=image, x=100, y=100, batch=batch)
sprite.scale = 8


@window.event
def on_draw():
    window.clear()
    batch.draw()


if __name__ == "__main__":
    pyglet.app.run()




# Tiger and later: need to set kWindowApplicationScaledAttribute for DPI independence?
from __future__ import annotations

import math
import warnings
from ctypes import byref, c_int32, c_void_p, string_at, cast, c_char_p, c_float
from typing import BinaryIO

import pyglet.image
from pyglet.font import base
from pyglet.font.base import Glyph, GlyphPosition
from pyglet.libs.darwin import CGFloat, cocoapy, kCTFontURLAttribute, cfnumber_to_number, \
    kCTFontWeightTrait
from pyglet.font.harfbuzz import harfbuzz_available, get_resource_from_ct_font, \
    get_harfbuzz_shaped_glyphs


cf = cocoapy.cf
ct = cocoapy.ct
quartz = cocoapy.quartz

UIFontWeightUltraLight = -0.8
UIFontWeightThin = -0.6
UIFontWeightLight = -0.4
UIFontWeightRegular = 0.0
UIFontWeightMedium = 0.23
UIFontWeightSemibold = 0.3
UIFontWeightBold = 0.4
UIFontWeightHeavy = 0.56
UIFontWeightBlack = 0.62

name_to_weight = {
    True: UIFontWeightBold,     # Bold as default for True
    False: UIFontWeightRegular,    # Regular for False
    None: UIFontWeightRegular,     # Regular if no weight provided
    "thin": UIFontWeightThin,
    "extralight": UIFontWeightUltraLight,
    "ultralight": UIFontWeightUltraLight,
    "light": UIFontWeightLight,
    "semilight": UIFontWeightLight,
    "normal": UIFontWeightRegular,
    "regular": UIFontWeightRegular,
    "medium": UIFontWeightMedium,
    "demibold": UIFontWeightSemibold,
    "semibold": UIFontWeightSemibold,
    "bold": UIFontWeightBold,
    "extrabold": UIFontWeightBold,
    "ultrabold": UIFontWeightBold,
    "black": UIFontWeightBlack,
    "heavy": UIFontWeightHeavy,
    "extrablack": UIFontWeightBlack,
}

name_to_stretch = {
    None: 1.0,
    False: 1.0,
    "undefined": 1.0,
    "ultracondensed": -0.4,
    "extracondensed": -0.3,
    "condensed": -0.2,
    "semicondensed": -0.1,
    "normal": 0.0,
    "medium": 0.0,
    "semiexpanded": 0.1,
    "expanded": 0.2,
    "extraexpanded": 0.3,
    "ultraexpanded": 0.4,
}


if harfbuzz_available():
    """Build the callbacks and information needed for Harfbuzz to work with CoreText Fonts.
    
    Getting the font data is not always reliable, and since no other way exists to
    retrieve the full font bytes from memory, we must construct callbacks for harfbuzz
    to retrieve the tag tables.
    """
    from pyglet.font.harfbuzz.harfbuzz_lib import hb_lib, hb_destroy_func_t, hb_reference_table_func_t, HB_MEMORY_MODE_READONLY

    def py_coretext_table_data_destroy(user_data: c_void_p):
        """Release the table resources once harfbuzz is done."""
        if user_data:
            cf.CFRelease(user_data)

    py_coretext_table_data_destroy_c = hb_destroy_func_t(py_coretext_table_data_destroy)

    @hb_reference_table_func_t
    def py_coretext_table_callback(face: c_void_p, tag: int, user_data: c_void_p):
        """This callback is invoked by HarfBuzz for each table it needs.

        user_data is a pointer to the CGFont.
        """
        # Use Quartz to get the table data for the given tag.
        table_data = quartz.CGFontCopyTableForTag(user_data, tag)
        if table_data is None:
            return None

        # Get the length and pointer to the raw data.
        length = cf.CFDataGetLength(table_data)
        data_ptr = cf.CFDataGetBytePtr(table_data)
        if not data_ptr:
            # Release the table_data and return empty blob.
            cf.CFRelease(table_data)
            return None

        # Create a blob that references this table data.
        data_ptr_char = cast(data_ptr, c_char_p)
        blob = hb_lib.hb_blob_create(data_ptr_char, length, HB_MEMORY_MODE_READONLY,
                                     table_data, py_coretext_table_data_destroy_c)
        return blob

    # Null callback, so harfbuzz cannot destroy our CGFont.
    _destroy_callback_null = cast(None, hb_destroy_func_t)

class QuartzGlyphRenderer(base.GlyphRenderer):
    font: QuartzFont

    def __init__(self, font: QuartzFont) -> None:
        super().__init__(font)
        self.font = font

    def render_index(self, glyph_index: int):
        ctFont = self.font.ctFont

        # Create an attributed string using text and font.
        # Determine the glyphs involved for the text (if any)
        # Get a bounding rectangle for glyphs in string.
        count = 1
        glyphs = (cocoapy.CGGlyph * count)(glyph_index)

        rect = ct.CTFontGetBoundingRectsForGlyphs(ctFont, 0, glyphs, None, count)

        # Get advance for all glyphs in string.
        advance = ct.CTFontGetAdvancesForGlyphs(ctFont, 0, glyphs, None, count)

        # Set image parameters:
        # We add 2 pixels to the bitmap width and height so that there will be a 1-pixel border
        # around the glyph image when it is placed in the texture atlas.  This prevents
        # weird artifacts from showing up around the edges of the rendered glyph textures.
        # We adjust the baseline and lsb of the glyph by 1 pixel accordingly.
        width = max(int(math.ceil(rect.size.width) + 2), 1)
        height = max(int(math.ceil(rect.size.height) + 2), 1)
        baseline = -int(math.floor(rect.origin.y)) + 1
        lsb = int(math.ceil(rect.origin.x)) - 1
        advance = int(round(advance))

        # Create bitmap context.
        bits_per_components = 8
        bytes_per_row = 4 * width
        colorSpace = c_void_p(quartz.CGColorSpaceCreateDeviceRGB())
        bitmap_context = c_void_p(quartz.CGBitmapContextCreate(
            None,
            width,
            height,
            bits_per_components,
            bytes_per_row,
            colorSpace,
            cocoapy.kCGImageAlphaPremultipliedLast))

        # Draw text to bitmap context.
        quartz.CGContextSetShouldAntialias(bitmap_context, pyglet.options.text_antialiasing)
        quartz.CGContextSetTextPosition(bitmap_context, -lsb, baseline)
        quartz.CGContextSetRGBFillColor(bitmap_context, 1, 1, 1, 1)
        quartz.CGContextSetFont(bitmap_context, ctFont)
        quartz.CGContextSetFontSize(bitmap_context, self.font.pixel_size)
        quartz.CGContextTranslateCTM(bitmap_context, 0, height)  # Move origin to top-left
        quartz.CGContextScaleCTM(bitmap_context, 1, -1)  # Flip vertically

        positions = (cocoapy.CGPoint * 1)(*[cocoapy.CGPoint(0, 0)])
        quartz.CTFontDrawGlyphs(ctFont, glyphs, positions, 1, bitmap_context)

        # Create an image to get the data out.
        image_ref = c_void_p(quartz.CGBitmapContextCreateImage(bitmap_context))

        bytes_per_row = quartz.CGImageGetBytesPerRow(image_ref)
        data_provider = c_void_p(quartz.CGImageGetDataProvider(image_ref))
        image_data = c_void_p(quartz.CGDataProviderCopyData(data_provider))
        buffer_size = cf.CFDataGetLength(image_data)
        buffer_ptr = cf.CFDataGetBytePtr(image_data)
        if buffer_ptr:
            buffer = string_at(buffer_ptr, buffer_size)

            quartz.CGImageRelease(image_ref)
            quartz.CGDataProviderRelease(image_data)
            cf.CFRelease(bitmap_context)
            cf.CFRelease(colorSpace)

            glyph_image = pyglet.image.ImageData(width, height, "RGBA", buffer, bytes_per_row)

            glyph = self.font.create_glyph(glyph_image)
            glyph.set_bearings(baseline, lsb, advance)

            return glyph

        raise Exception("CG Image buffer could not be read.")

    def render(self, text: str) -> base.Glyph:
        # Using CTLineDraw seems to be the only way to make sure that the text
        # is drawn with the specified font when that font is a graphics font loaded from
        # memory.  For whatever reason, [NSAttributedString drawAtPoint:] ignores
        # the graphics font if it not registered with the font manager.
        # So we just use CTLineDraw for both graphics fonts and installed fonts.

        ctFont = self.font.ctFont

        # Create an attributed string using text and font.
        attributes = c_void_p(
            cf.CFDictionaryCreateMutable(None, 2, cf.kCFTypeDictionaryKeyCallBacks, cf.kCFTypeDictionaryValueCallBacks)
        )
        cf.CFDictionaryAddValue(attributes, cocoapy.kCTFontAttributeName, ctFont)
        cf.CFDictionaryAddValue(attributes, cocoapy.kCTForegroundColorFromContextAttributeName , cocoapy.kCFBooleanTrue)
        cf_str = cocoapy.CFSTR(text)
        string = c_void_p(cf.CFAttributedStringCreate(None, cf_str, attributes))

        # Create a CTLine object to render the string.
        line = c_void_p(ct.CTLineCreateWithAttributedString(string))
        cf.CFRelease(string)
        cf.CFRelease(attributes)
        cf.CFRelease(cf_str)

        # Determine the glyphs involved for the text (if any)
        count = len(text)
        chars = (cocoapy.UniChar * count)(*list(map(ord, str(text))))
        glyphs = (cocoapy.CGGlyph * count)()
        ct.CTFontGetGlyphsForCharacters(ctFont, chars, glyphs, count)

        # If a glyph is returned as 0, it does not exist in the current font.
        if glyphs[0] == 0:
            # Use the typographic bounds instead for the placements.
            # This seems to have some sort of fallback information in the bounds.
            ascent, descent = CGFloat(), CGFloat()
            advance = width = int(ct.CTLineGetTypographicBounds(line, byref(ascent), byref(descent), None))
            height = int(ascent.value + descent.value)
            lsb = 0
            baseline = descent.value
        else:
            # Get a bounding rectangle for glyphs in string.
            rect = ct.CTFontGetBoundingRectsForGlyphs(ctFont, 0, glyphs, None, count)

            # Get advance for all glyphs in string.
            advance = ct.CTFontGetAdvancesForGlyphs(ctFont, 0, glyphs, None, count)

            # Set image parameters:
            # We add 2 pixels to the bitmap width and height so that there will be a 1-pixel border
            # around the glyph image when it is placed in the texture atlas.  This prevents
            # weird artifacts from showing up around the edges of the rendered glyph textures.
            # We adjust the baseline and lsb of the glyph by 1 pixel accordingly.

            width = max(int(math.ceil(rect.size.width) + 2), 1)
            height = max(int(math.ceil(rect.size.height) + 2), 1)
            baseline = -int(math.floor(rect.origin.y)) + 1
            lsb = int(math.ceil(rect.origin.x)) - 1
            advance = int(round(advance))

        # Create bitmap context.
        bits_per_components = 8
        bytes_per_row = 4 * width
        colorSpace = c_void_p(quartz.CGColorSpaceCreateDeviceRGB())
        bitmap_context = c_void_p(quartz.CGBitmapContextCreate(
            None,
            width,
            height,
            bits_per_components,
            bytes_per_row,
            colorSpace,
            cocoapy.kCGImageAlphaPremultipliedLast))

        # Draw text to bitmap context.
        quartz.CGContextSetShouldAntialias(bitmap_context, pyglet.options.text_antialiasing)
        quartz.CGContextSetTextPosition(bitmap_context, -lsb, baseline)
        quartz.CGContextSetRGBFillColor(bitmap_context, 1, 1, 1, 1)  # Render white for multiplying.
        quartz.CGContextTranslateCTM(bitmap_context, 0, height)  # Move origin to top-left
        quartz.CGContextScaleCTM(bitmap_context, 1, -1)  # Flip vertically

        ct.CTLineDraw(line, bitmap_context)
        cf.CFRelease(line)
        # Create an image to get the data out.
        image_ref = c_void_p(quartz.CGBitmapContextCreateImage(bitmap_context))

        bytes_per_row = quartz.CGImageGetBytesPerRow(image_ref)
        data_provider = c_void_p(quartz.CGImageGetDataProvider(image_ref))
        image_data = c_void_p(quartz.CGDataProviderCopyData(data_provider))
        buffer_size = cf.CFDataGetLength(image_data)
        buffer_ptr = cf.CFDataGetBytePtr(image_data)
        if buffer_ptr:
            buffer = string_at(buffer_ptr, buffer_size)

            quartz.CGImageRelease(image_ref)
            quartz.CGDataProviderRelease(image_data)
            cf.CFRelease(bitmap_context)
            cf.CFRelease(colorSpace)

            glyph_image = pyglet.image.ImageData(width, height, "RGBA", buffer, bytes_per_row)

            glyph = self.font.create_glyph(glyph_image)
            glyph.set_bearings(baseline, lsb, advance)

            return glyph

        raise Exception("CG Image buffer could not be read.")


class QuartzFont(base.Font):
    glyph_renderer_class: type[base.GlyphRenderer] = QuartzGlyphRenderer
    _loaded_CGFont_table: dict[str, dict[int, tuple[c_void_p, bytes]]] = {}

    def __init__(self, name: str, size: float, weight: str = "normal", italic: bool = False, stretch: bool = False,
                 dpi: int | None = None) -> None:

        super().__init__()

        name = name or "Helvetica"

        self.dpi = dpi or 96
        self.size = size
        self.pixel_size = size * dpi / 72.0
        self.italic = italic
        self.stretch = stretch
        self.weight = weight

        if isinstance(weight, str):
            self.weight_value = name_to_weight[weight]
        elif weight is True:
            self.weight_value = name_to_weight["bold"]
        else:
            self.weight_value = None

        self.italic = italic

        # Construct traits value.
        traits = 0
        if italic:
            traits |= cocoapy.kCTFontItalicTrait

        if isinstance(stretch, str):
            self.stretch_value = name_to_stretch[stretch]
        else:
            self.stretch_value = None

        name = str(name)
        self.traits = traits
        # First see if we can find an appropriate font from our table of loaded fonts.
        result = self._lookup_font_with_family_and_traits(name, traits)
        if result:
            cgFont = result[0]
            # Use cgFont from table to create a CTFont object with the specified size.
            self.ctFont = c_void_p(ct.CTFontCreateWithGraphicsFont(cgFont, self.pixel_size, None, None))
        else:
            # Create a font descriptor for given name and traits and use it to create font.
            descriptor = self._create_font_descriptor(name, traits, self.weight_value, self.stretch_value)
            self.ctFont = c_void_p(ct.CTFontCreateWithFontDescriptor(descriptor, self.pixel_size, None))
            cf.CFRelease(descriptor)
            assert self.ctFont, "Couldn't load font: " + name

        string = c_void_p(ct.CTFontCopyFamilyName(self.ctFont))
        self._family_name = str(cocoapy.cfstring_to_string(string))
        cf.CFRelease(string)

        self.ascent = int(math.ceil(ct.CTFontGetAscent(self.ctFont)))
        self.descent = -int(math.ceil(ct.CTFontGetDescent(self.ctFont)))

        if pyglet.options.text_shaping == 'harfbuzz' and harfbuzz_available():
            self._cg_font = None
            self.hb_resource = get_resource_from_ct_font(self)


    def _get_hb_face(self):
        assert self._cg_font is None

        # Create a CGFont from the CTFont for the face.
        self._cg_font = quartz.CTFontCopyGraphicsFont(self.ctFont, None)
        if self._cg_font is None:
            raise ValueError("Could not get CGFont from CTFont")

        # Create the HarfBuzz face using our table callback.
        return hb_lib.hb_face_create_for_tables(py_coretext_table_callback, self._cg_font, _destroy_callback_null)

    def get_font_data(self) -> bytes:
        """Get the font file in bytes if possible.

        Unfortunately CoreText doesn't have a good way to retrieve directly from a font object. Attempt to get the
        filename from the system. If the filename is unknown, it most likely was loaded from memory. In which case,
        the data was added and cached at some point with `add_font_data`.
        """
        filename = self.filename
        if filename == "Unknown":
            result = self._lookup_font_with_family_and_traits(self.name, self.traits)
            if result:
                _, data = result
            else:
                raise Exception("Couldn't load font data by name and traits. Report as a bug with the font file.")
        else:
            with open(filename, "rb") as f:
                data = f.read()

        return data

    @property
    def filename(self) -> str:
        descriptor = self._create_font_descriptor(self.name, self.traits, self.weight_value, self.stretch_value)
        ref = c_void_p(ct.CTFontDescriptorCopyAttribute(descriptor, kCTFontURLAttribute))
        if ref:
            url = cocoapy.ObjCInstance(ref)  # NSURL
            filepath = url.fileSystemRepresentation().decode()
            cf.CFRelease(ref)
            cf.CFRelease(descriptor)
            return filepath

        cf.CFRelease(descriptor)
        return "Unknown"

    @property
    def name(self) -> str:
        return self._family_name

    def __del__(self) -> None:
        cf.CFRelease(self.ctFont)

    def _lookup_font_with_family_and_traits(self, family: str, traits: int) -> tuple[c_void_p, bytes] | None:
        # This method searches the _loaded_CGFont_table to find a loaded
        # font of the given family with the desired traits.  If it can't find
        # anything with the exact traits, it tries to fall back to whatever
        # we have loaded that's close.  If it can't find anything in the
        # given family at all, it returns None.

        # Check if we've loaded the font with the specified family.
        if family not in self._loaded_CGFont_table:
            return None
        # Grab a dictionary of all fonts in the family, keyed by traits.
        fonts = self._loaded_CGFont_table[family]
        if not fonts:
            return None
        # Return font with desired traits if it is available.
        if traits in fonts:
            return fonts[traits]
        # Otherwise try to find a font with some of the traits.
        for (t, f) in fonts.items():
            if traits & t:
                return f
        # Otherwise try to return a regular font.
        if 0 in fonts:
            return fonts[0]
        # Otherwise return whatever we have.
        return list(fonts.values())[0]

    def _create_font_descriptor(self, family_name: str, traits: int,
                                weight: float | None = None,
                                stretch: float | None = None) -> c_void_p:
        # Create an attribute dictionary.
        attributes = c_void_p(
            cf.CFDictionaryCreateMutable(None, 0, cf.kCFTypeDictionaryKeyCallBacks, cf.kCFTypeDictionaryValueCallBacks))
        # Add family name to attributes.
        cfname = cocoapy.CFSTR(family_name)
        cf.CFDictionaryAddValue(attributes, cocoapy.kCTFontFamilyNameAttribute, cfname)
        cf.CFRelease(cfname)

        # Construct a CFNumber to represent the traits.
        itraits = c_int32(traits)
        symTraits = c_void_p(cf.CFNumberCreate(None, cocoapy.kCFNumberSInt32Type, byref(itraits)))
        if symTraits:
            # Construct a dictionary to hold the traits values.
            traitsDict = c_void_p(cf.CFDictionaryCreateMutable(None, 0, cf.kCFTypeDictionaryKeyCallBacks,
                                                               cf.kCFTypeDictionaryValueCallBacks))
            if traitsDict:
                if weight is not None:
                    weight_value = c_float(weight)
                    cfWeight = c_void_p(cf.CFNumberCreate(None, cocoapy.kCFNumberFloatType, byref(weight_value)))
                    if cfWeight:
                        cf.CFDictionaryAddValue(traitsDict, cocoapy.kCTFontWeightTrait, cfWeight)
                        cf.CFRelease(cfWeight)

                if stretch is not None:
                    stretch_value = c_float(stretch)
                    cfWidth = c_void_p(cf.CFNumberCreate(None, cocoapy.kCFNumberFloatType, byref(stretch_value)))
                    if cfWidth:
                        cf.CFDictionaryAddValue(traitsDict, cocoapy.kCTFontWidthTrait, cfWidth)
                        cf.CFRelease(cfWidth)

                # Add CFNumber traits to traits dictionary.
                cf.CFDictionaryAddValue(traitsDict, cocoapy.kCTFontSymbolicTrait, symTraits)
                # Add traits dictionary to attributes.
                cf.CFDictionaryAddValue(attributes, cocoapy.kCTFontTraitsAttribute, traitsDict)
                cf.CFRelease(traitsDict)

            cf.CFRelease(symTraits)
        # Create font descriptor with attributes.
        descriptor = c_void_p(ct.CTFontDescriptorCreateWithAttributes(attributes))
        cf.CFRelease(attributes)
        return descriptor

    @classmethod
    def have_font(cls: type[QuartzFont], name: str) -> bool:
        name = str(name)
        if name in cls._loaded_CGFont_table:
            return True

        # Query CoreText to see if the font exists.
        descriptor = ct.CTFontDescriptorCreateWithNameAndSize(cocoapy.CFSTR(name), 0.0)
        matched = ct.CTFontDescriptorCreateMatchingFontDescriptor(descriptor, None)

        exists = True if matched else False

        if descriptor:
            cf.CFRelease(descriptor)

        if matched:
            cf.CFRelease(matched)

        return exists

    @classmethod
    def add_font_data(cls: type[QuartzFont], data: BinaryIO) -> None:
        # Create a cgFont with the data.  There doesn't seem to be a way to
        # register a font loaded from memory such that the operating system will
        # find it later.  So instead we just store the cgFont in a table where
        # it can be found by our __init__ method.
        # Note that the iOS CTFontManager *is* able to register graphics fonts,
        # however this method is missing from CTFontManager on MacOS 10.6
        dataRef = c_void_p(cf.CFDataCreate(None, data, len(data)))
        provider = c_void_p(quartz.CGDataProviderCreateWithCFData(dataRef))
        cgFont = c_void_p(quartz.CGFontCreateWithDataProvider(provider))

        cf.CFRelease(dataRef)
        quartz.CGDataProviderRelease(provider)

        # Create a template CTFont from the graphics font so that we can get font info.
        ctFont = c_void_p(ct.CTFontCreateWithGraphicsFont(cgFont, 1, None, None))

        # Get info about the font to use as key in our font table.
        string = c_void_p(ct.CTFontCopyFamilyName(ctFont))
        familyName = str(cocoapy.cfstring_to_string(string))
        cf.CFRelease(string)

        string = c_void_p(ct.CTFontCopyFullName(ctFont))
        fullName = str(cocoapy.cfstring_to_string(string))
        cf.CFRelease(string)

        traits = ct.CTFontGetSymbolicTraits(ctFont)
        cf.CFRelease(ctFont)

        # Store font in table. We store it under both its family name and its
        # full name, since its not always clear which one will be looked up.
        if familyName not in cls._loaded_CGFont_table:
            cls._loaded_CGFont_table[familyName] = {}
        cls._loaded_CGFont_table[familyName][traits] = (cgFont, data)

        if fullName not in cls._loaded_CGFont_table:
            cls._loaded_CGFont_table[fullName] = {}
        cls._loaded_CGFont_table[fullName][traits] = (cgFont, data)

    def get_text_size(self, text: str) -> tuple[int, int]:
        ctFont = self.ctFont

        attributes = c_void_p(
            cf.CFDictionaryCreateMutable(None, 1, cf.kCFTypeDictionaryKeyCallBacks, cf.kCFTypeDictionaryValueCallBacks)
        )
        cf.CFDictionaryAddValue(attributes, cocoapy.kCTFontAttributeName, ctFont)
        cf_str = cocoapy.CFSTR(text)
        string = c_void_p(cf.CFAttributedStringCreate(None, cf_str, attributes))

        line = c_void_p(ct.CTLineCreateWithAttributedString(string))

        ascent, descent = CGFloat(), CGFloat()
        width = ct.CTLineGetTypographicBounds(line, byref(ascent), byref(descent), None)
        height = ascent.value + descent.value

        cf.CFRelease(string)
        cf.CFRelease(attributes)
        cf.CFRelease(line)
        cf.CFRelease(cf_str)
        return round(width), round(height)

    def _get_font_weight(self):
        traits = ct.CTFontCopyTraits(self.ctFont)
        font_weight = cfnumber_to_number(c_void_p(cf.CFDictionaryGetValue(traits, kCTFontWeightTrait)))
        cf.CFRelease(traits)
        return font_weight

    def _get_font_stretch(self):
        traits = ct.CTFontCopyTraits(self.ctFont)
        font_width = cfnumber_to_number(c_void_p(cf.CFDictionaryGetValue(traits, cocoapy.kCTFontWidthTrait)))
        cf.CFRelease(traits)
        return font_width

    def render_glyph_indices(self, indices: list[int]) -> None:
        # Process any glyphs that have not been rendered.
        self._initialize_renderer()

        missing = set()
        for glyph_indice in set(indices):
            if glyph_indice not in self.glyphs:
                missing.add(glyph_indice)

        # Missing glyphs, get their info.
        for glyph_indice in missing:
            self.glyphs[glyph_indice] = self._glyph_renderer.render_index(glyph_indice)

    def get_glyphs(self, text: str) -> tuple[list[Glyph], list[GlyphPosition]]:
        """Create and return a list of Glyphs for `text`.

        If any characters do not have a known glyph representation in this
        font, a substitution will be made.

        Args:
            text:
                Text to render.
        """
        self._initialize_renderer()

        if pyglet.options.text_shaping == "harfbuzz" and harfbuzz_available():
            return get_harfbuzz_shaped_glyphs(self, text)

        glyphs = []  # glyphs that are committed.
        offsets = []
        for c in base.get_grapheme_clusters(str(text)):
            if c == "\t":
                c = " "  # noqa: PLW2901
            if c not in self.glyphs:
                self.glyphs[c] = self._glyph_renderer.render(c)
            glyphs.append(self.glyphs[c])
            offsets.append(base.GlyphPosition(0, 0, 0, 0))

        return glyphs, offsets


"""
Usage

    run_test_suite.py [samples_dir]

Plays media samples with the pyglet media_player, recording debug information
for each sample played and writing reports about the data captured.

Arguments
    samples_dir: directory with the media samples to play

If no sample_dir is provided the samples to play will come from the active
session (see cmd configure new)

If the samples_dir is provided, all files within will be played except for the
ones with extension ".dbg", ".htm", ".html", ".json", ".log", ".pkl", ".py",
".txt" ; subdirectories of samples_dir will not be explored. 

Output files will be into
    samples_dir/testrun/dbg : binary capture of media_player events
    samples_dir/testrun/reports : human readable reports

(with testrun provided by the active session if sample_dir was not provided)
"""

import os
import sys

import fs
from pyglet.media import instrumentation as ins
import playmany
import retry_crashed
import summarize


def main(samples_dir):
    if samples_dir is None:
        # get from active session
        pathserv = fs.get_path_info_for_active_session()
    else:
        # create a session with a default name, provide a pathserv
        session = None
        for i in range(100):
            name = "testrun_%02d" % i
            if not os.path.exists(os.path.join(samples_dir, name)):
                session = name
                pathserv = fs.new_session_for_samples_dir(session, samples_dir)
                break
        if session is None:
            s = os.path.join(samples_dir, "testrun_n")
            print("*** Error, failed to create output dir in the form %s , for any n<100." % s)
            sys.exit(1)

    print(f"Results will be write to directory: {pathserv.session_dir}")

    playmany.play_many(pathserv)

    retry_crashed.retry_crashed(pathserv)

    count_bads = ins.CountBads()
    summarize.summarize(pathserv, count_bads)

    # protect raw data and reports
    modify = {
        "protect_raw_data": True,
        "protect_reports": True,
        }
    fs.update_configuration(pathserv, modify)
    
    print("*** done ***") 


def usage():
    print(__doc__)
    sys.exit(1)


def sysargs_to_mainargs():
    """builds main args from sys.argv"""
    if len(sys.argv) > 1:
        for i in range(1):
            if sys.argv[1].startswith("--"):
                a = sys.argv.pop(1)
                if a.startswith("--help"):
                    usage()
                else:
                    print("Error unknown option:", a)
                    usage()

    if len(sys.argv) > 1:
        samples_dir = sys.argv.pop(1)
    else:
        samples_dir = None

    if len(sys.argv) > 1:
        print("Error, unexpected extra params: %s ..." % sys.argv[1])
        usage()

    return samples_dir 

if __name__ == '__main__':
    samples_dir = sysargs_to_mainargs()
    main(samples_dir)



"""Test creation of all Layout classes"""
import random
import itertools


import pytest

from pyglet.text import document
from pyglet.text import layout

WIDTH = 500
HEIGHT = 100
X = random.randint(0, 900)
Y = random.randint(0, 600)
Z = random.randint(-10, 10)


# Create combination of all Layout and Document types
document_classes = [document.UnformattedDocument, document.FormattedDocument]
layout_classes = [layout.TextLayout, layout.ScrollableTextLayout, layout.IncrementalTextLayout]
all_combinations = list(itertools.product(document_classes, layout_classes))


@pytest.mark.parametrize('doctype, layouttype', all_combinations)
def test_layout_creation_keyword(doctype, layouttype):
    _doc = doctype("This is a test")
    _layout = layouttype(document=_doc, x=X, y=Y, z=Z, width=WIDTH, height=HEIGHT)
    assert _layout.x == X
    assert _layout.y == Y
    assert _layout.z == Z
    assert _layout.width == WIDTH
    assert _layout.height == HEIGHT
    assert _layout.position == (X, Y, Z)


@pytest.mark.parametrize('doctype, layouttype', all_combinations)
def test_layout_creation_positional(doctype, layouttype):
    _doc = doctype("This is a test")
    _layout = layouttype(_doc, X, Y, Z, WIDTH, HEIGHT)
    # Make sure the arguments were in order:
    assert _layout.x == X
    assert _layout.y == Y
    assert _layout.z == Z
    assert _layout.width == WIDTH
    assert _layout.height == HEIGHT
    assert _layout.position == (X, Y, Z)



"""Encoder and decoder for PNG files, using PyPNG (png.py).
"""

import array
import itertools

from pyglet.image import ImageData
from pyglet.image.codecs import ImageDecoder, ImageEncoder, ImageDecodeException

import pyglet.extlibs.png as pypng


class PNGImageDecoder(ImageDecoder):
    def get_file_extensions(self):
        return ['.png']

    def decode(self, filename, file):
        if not file:
            file = open(filename, 'rb')

        try:
            reader = pypng.Reader(file=file)
            width, height, pixels, metadata = reader.asDirect()
        except Exception as e:
            raise ImageDecodeException('PyPNG cannot read %r: %s' % (filename or file, e))

        if metadata['greyscale']:
            if metadata['alpha']:
                fmt = 'LA'
            else:
                fmt = 'L'
        else:
            if metadata['alpha']:
                fmt = 'RGBA'
            else:
                fmt = 'RGB'
        pitch = len(fmt) * width

        pixels = array.array('BH'[metadata['bitdepth'] > 8], itertools.chain(*pixels))
        return ImageData(width, height, fmt, pixels.tobytes(), -pitch)


class PNGImageEncoder(ImageEncoder):
    def get_file_extensions(self):
        return ['.png']

    def encode(self, image, filename, file):
        image = image.get_image_data()

        has_alpha = 'A' in image.format
        greyscale = len(image.format) < 3
        if has_alpha:
            if greyscale:
                image.format = 'LA'
            else:
                image.format = 'RGBA'
        else:
            if greyscale:
                image.format = 'L'
            else:
                image.format = 'RGB'

        image.pitch = -(image.width * len(image.format))

        writer = pypng.Writer(image.width, image.height, greyscale=greyscale, alpha=has_alpha)

        data = array.array('B')
        data.frombytes(image.get_data(image.format, image.pitch))

        writer.write_array(file, data)


def get_decoders():
    return [PNGImageDecoder()]


def get_encoders():
    return [PNGImageEncoder()]



"""
Responsibilities

Containers with helper functions to analyze media_player recording(s)

Commands 'compare' and 'summarize' uses them to build its reports.

"""

import os

import fs


class OverallDefectsCounter:
    """Helper class to count defects over a range of media player sample plays

    Collaborates with a caller, which determines which samples to process and
    which samples to skip.

    The main result is a dict of defect: count over all suitable samples

    Additionally, classifies the samples in the disjoint sets:
        crashed_samples: samples that crashed
        no_dbg_samples: samples with no media_player state recording
        skipped_samples: samples that the caller excludes from the overall counts
        perfect_play_samples: non-skipped samples that played perfectly
        counters_non_perfect_play_samples: A dict of
            sample: <counters for each defect type found in recording>
            for each sample with dbg, no crash, no perfect, not skipped.

    The overall defect counters only uses the samples in the last group.

    Used to:
        - calculate a session quality score (crashes, defects per sample)
        - decide which of two session behaves better

    Notice that the more samples are discarded, the less significant will be
    the score; a good report will need to consider that.

    design notes
        - this is a bit overcomplicated for single session score, but
          probably the minimum while comparing two sessions.
    """

    def __init__(self, pathserv, fn_count_bads):
        self.pathserv = pathserv
        self.fn_count_bads = fn_count_bads
        self.crashed_samples = set()
        self.no_dbg_samples = set()
        self.perfect_play_samples = set()
        self.counters_non_perfect_play_samples = {}
        self.skipped_samples = set()
        self.overall_counters = dict()
        self.overall_counters["scheduling_in_past"] = 0
        self.last_count = None

    def count_defects(self, sample):
        dbg_file = self.pathserv.dbg_filename(sample)
        if not os.path.isfile(dbg_file):
            self.no_dbg_samples.add(sample)
            classifies_as = "no_dbg"
            counters = None
        else:
            log_entries = fs.pickle_load(dbg_file)
            counters = self.fn_count_bads(log_entries)
            if counters["crash"]:
                self.crashed_samples.add(sample)
                classifies_as = "crash"
            else:
                perfect = all((counters[k] == 0 for k in counters))
                classifies_as = "perfect" if perfect else "non_perfect"
        self.last_count = sample, classifies_as, counters
        return classifies_as

    def add_count(self, skip=False):
        sample, classifies_as, counters = self.last_count
        self.last_count = None
        if classifies_as in {"no_dbg", "crash"}:
            return
        if skip:
            self.skipped_samples.add(sample)
            return
        if classifies_as == "perfect":
            self.perfect_play_samples.add(sample)
        else:
            self.counters_non_perfect_play_samples[sample] = counters
            for k in counters:
                self.overall_counters[k] = self.overall_counters.get(k, 0) + counters[k]

    def sum_overall(self):
        total = 0
        for k in self.overall_counters:
            total += self.overall_counters[k]
        return total

    def total_relevant_samples(self):
        total = (len(self.perfect_play_samples) +
                 len(self.counters_non_perfect_play_samples))
        return total

    def defects_per_sample(self):
        return self.sum_overall() / self.total_relevant_samples()


def single_session_overall_defect_counters(pathserv, fn_count_bads):
    overall = OverallDefectsCounter(pathserv, fn_count_bads)
    samples = {e[0] for e in pathserv.session_playlist_generator()}
    for sample in samples:
        overall.count_defects(sample)
        overall.add_count()
    return overall


class CollectSessionComparisonData:
    """Collects data to compare two sessions

    Constructs the sets samples, other_samples, common_samples, no_dbg_samples,
    crashed_samples, compared_samples

    For each session calculates an OverallDefectsCounter instance over the set
    compared_samples; results stored in overall_counters and
    overall_counters_other.
    """
    def __init__(self, pathserv, pathserv_other, fn_count_bads):
        self.pathserv = pathserv
        self.pathserv_other = pathserv_other

        # intersection playlist
        samples = {e for e in fs.load_session_playlist(pathserv)}
        other_samples = {e for e in fs.load_session_playlist(pathserv_other)}
        common_samples = samples & other_samples

        # count defects, only for samples with no crash in both sessions
        no_dbg_samples = set()
        crashed_samples = set()
        compared_samples = set()
        cls = OverallDefectsCounter
        overall_counters = cls(pathserv, fn_count_bads)
        overall_counters_o = cls(pathserv_other, fn_count_bads)
        for sample in common_samples:
            perfect, counters = overall_counters.count_defects(sample)
            perfect_o, counters_o = overall_counters_o.count_defects(sample)
            if counters["no_dbg"] or counters_o["no_dbg"]:
                no_dbg_samples.add(sample)
                continue
            if counters["crash"] or counters_o["crash"]:
                crashed_samples.add(sample)
                continue
            compared_samples.add(sample)
            overall_counters.count_defects(sample)
            overall_counters_o.count_defects(sample)

        self.no_dbg_samples = no_dbg_samples
        self.samples = samples
        self.other_samples = other_samples
        self.common_samples = common_samples
        self.crashed_samples = crashed_samples
        self.compared_samples = compared_samples
        self.overall_counters = overall_counters
        self.overall_counters_other = overall_counters_o

    def confidence_guess(self):
        """estimates how meaningful the defects per sample are

        The value will be between 1(most meaningful) and 0(useless).

        It follows the heuristic 'the more samples are discarded, the less
        meaning will be the conclusions draw'
        """
        maxz = max(len(self.samples), len(self.other_samples))
        c = len(self.compared_samples) / maxz if maxz > 0 else 0
        return c



import os
import ctypes
import unittest
from unittest import mock


import pyglet
from pyglet.media.drivers.base import MediaEvent
from pyglet.media.codecs.base import *



def _bytes_from_audiodata(audio_data):
    return memoryview(audio_data.data).tobytes()


class AudioFormatTestCase(unittest.TestCase):
    def test_equality_true(self):
        af1 = AudioFormat(2, 8, 44100)
        af2 = AudioFormat(2, 8, 44100)
        self.assertEqual(af1, af2)

    def test_equality_false(self):
        channels = [1, 2]
        sample_sizes = [8, 16]
        sample_rates = [11025, 22050, 44100]

        formats = [AudioFormat(c, s, r) for c in channels for s in sample_sizes for r in sample_rates]
        while formats:
            a = formats.pop()
            for b in formats:
                self.assertNotEqual(a, b)

    def test_bytes_per(self):
        af1 = AudioFormat(1, 8, 22050)
        af2 = AudioFormat(2, 16, 44100)

        self.assertEqual(af1.bytes_per_frame, 1)
        self.assertEqual(af1.bytes_per_second, 22050)

        self.assertEqual(af2.bytes_per_frame, 4)
        self.assertEqual(af2.bytes_per_second, 176400)

    def test_alignment(self):
        af = AudioFormat(2, 16, 44100)

        self.assertEqual(af.align(2049), 2048)
        self.assertEqual(af.align(2048), 2048)
        self.assertEqual(af.align(2047), 2044)
        self.assertEqual(af.align(0), 0)
        self.assertEqual(af.align(-1), -4)
        self.assertEqual(af.align_ceil(2049), 2052)
        self.assertEqual(af.align_ceil(2048), 2048)
        self.assertEqual(af.align_ceil(2047), 2048)
        self.assertEqual(af.align_ceil(0), 0)
        self.assertEqual(af.align_ceil(-1), 0)

    def test_repr(self):
        af1 = AudioFormat(1, 8, 22050)
        self.assertEqual(repr(af1), 'AudioFormat(channels=1, sample_size=8, sample_rate=22050)')

        af2 = AudioFormat(2, 16, 44100)
        self.assertEqual(repr(af2), 'AudioFormat(channels=2, sample_size=16, sample_rate=44100)')


class AudioDataTestCase(unittest.TestCase):
    def generate_random_string_data(self, length):
        return os.urandom(length)

    def create_string_buffer(self, data):
        return ctypes.create_string_buffer(data)

    def test_pointer_bytes(self):
        source_bytes = self.generate_random_string_data(2048)
        assert len(source_bytes) == 2048

        audio_data = AudioData(source_bytes, 2048)

        self.assertIsInstance(audio_data.pointer, int)
        self.assertGreater(audio_data.pointer, 0)

        retrieved_bytes = ctypes.string_at(audio_data.pointer, audio_data.length)
        self.assertEqual(retrieved_bytes, source_bytes)

    def test_buffer_protocol(self):
        d0 = AudioData(b"01234567", 8)
        d1 = AudioData((ctypes.c_char * 24)(), 24)
        d2 = AudioData(bytearray(64), 64)

        self.assertEqual(_bytes_from_audiodata(d0), b"01234567")
        self.assertEqual(_bytes_from_audiodata(d1), b"\x00" * 24)
        self.assertEqual(_bytes_from_audiodata(d2), b"\x00" * 64)


class SourceTestCase(unittest.TestCase):
    @mock.patch('pyglet.media.player.Player')
    def test_play(self, player_mock):
        source = Source()
        returned_player = source.play()

        self.assertIsNotNone(returned_player)
        returned_player.play.assert_called_once_with()
        returned_player.queue.assert_called_once_with(source)

    @mock.patch('pyglet.media.codecs.base.Source.get_next_video_timestamp')
    @mock.patch('pyglet.media.codecs.base.Source.get_next_video_frame')
    def test_get_animation(self, mock_get_next_video_frame, mock_get_next_video_timestamp):
        def _next_timestamp():
            if _next_timestamp.timestamp < 100:
                _next_timestamp.timestamp += 1
                return float(_next_timestamp.timestamp / 10)
        def _next_frame():
            return _next_timestamp.timestamp
        _next_timestamp.timestamp = 0
        mock_get_next_video_frame.side_effect = _next_frame
        mock_get_next_video_timestamp.side_effect = _next_timestamp

        source = Source()
        source.video_format = VideoFormat(800, 600)

        animation = source.get_animation()
        self.assertIsNotNone(animation)
        self.assertIsInstance(animation, pyglet.image.Animation)
        self.assertEqual(len(animation.frames), 100)
        for frame in animation.frames:
            self.assertAlmostEqual(frame.duration, 0.1)

    @unittest.skip('pyglet.image.Animation does not like getting an empty list of frames.')
    def test_get_animation_no_video(self):
        source = Source()
        animation = source.get_animation()

        self.assertIsNotNone(animation)
        self.assertIsInstance(animation, pyglet.image.Animation)
        self.assertEqual(len(animation.frames), 0)

    def test_can_queue_only_once(self):
        source = StreamingSource()
        self.assertFalse(source.is_player_source)

        ret = source.get_queue_source()
        self.assertTrue(ret.is_player_source)
        self.assertTrue(source.is_player_source)

        with self.assertRaises(MediaException):
            source.get_queue_source()

        ret.is_player_source = False
        self.assertFalse(ret.is_player_source)
        self.assertFalse(source.is_player_source)

        ret = source.get_queue_source()
        with self.assertRaises(MediaException):
            ret.get_queue_source()

    def test_queue_source(self):
        source = StreamingSource()
        qsource = source.get_queue_source()
        self.assertIs(qsource, source)
        self.assertTrue(qsource.is_player_source)


class StaticSourceTestCase(unittest.TestCase):
    def create_valid_mock_source(self, bitrate=8, channels=1):
        self.mock_source = mock.MagicMock()
        self.mock_queue_source = self.mock_source.get_queue_source.return_value

        byte_rate = bitrate >> 3
        self.mock_data = [b'a'*22050*byte_rate*channels,
                          b'b'*22050*byte_rate*channels,
                          b'c'*11025*byte_rate*channels]
        self.mock_data_length = sum(map(len, self.mock_data))
        self.mock_audio_data = b''.join(self.mock_data)
        def _get_audio_data(_):
            if not self.mock_data:
                return None
            data = self.mock_data.pop(0)
            return AudioData(data, len(data), 0.0, 1.0, ())
        self.mock_queue_source.get_audio_data.side_effect = _get_audio_data

        self.mock_queue_source.audio_format = AudioFormat(channels, bitrate, 11025)
        self.mock_queue_source.video_format = None
        self.mock_source.audio_format = AudioFormat(channels, bitrate, 11025)
        self.mock_source.video_format = None


    def test_reads_all_data_on_init(self):
        self.create_valid_mock_source()
        static_source = StaticSource(self.mock_source)

        self.assertEqual(len(self.mock_data), 0, 'All audio data should be read')
        self.assertEqual(self.mock_queue_source.get_audio_data.call_count, 4, 'Data should be retrieved until empty')

        # Try to read all data plus more, more should be ignored
        returned_audio_data = static_source.get_queue_source().get_audio_data(len(self.mock_audio_data) + 1024)
        self.assertEqual(_bytes_from_audiodata(returned_audio_data), self.mock_audio_data, 'All data from the mock should be returned')
        self.assertAlmostEqual(returned_audio_data.duration, 5.0)

    def test_video_not_supported(self):
        self.create_valid_mock_source()
        self.mock_queue_source.video_format = VideoFormat(800, 600)

        with self.assertRaises(NotImplementedError):
            static_source = StaticSource(self.mock_source)

    def test_seek(self):
        self.create_valid_mock_source()
        static_source = StaticSource(self.mock_source)

        queue_source = static_source.get_queue_source()
        queue_source.seek(1.0)
        returned_audio_data = queue_source.get_audio_data(len(self.mock_audio_data))
        self.assertAlmostEqual(returned_audio_data.duration, 4.0)
        self.assertEqual(returned_audio_data.length, len(self.mock_audio_data)-11025)
        self.assertEqual(_bytes_from_audiodata(returned_audio_data), self.mock_audio_data[11025:], 'Should have seeked past 1 second')

    def test_multiple_queued(self):
        self.create_valid_mock_source()
        static_source = StaticSource(self.mock_source)

        # Check that seeking and consuming on queued instances does not affect others
        queue_source1 = static_source.get_queue_source()
        queue_source1.seek(1.0)

        queue_source2 = static_source.get_queue_source()

        returned_audio_data = queue_source2.get_audio_data(len(self.mock_audio_data))
        self.assertAlmostEqual(returned_audio_data.duration, 5.0)
        self.assertEqual(returned_audio_data.length, len(self.mock_audio_data), 'Should contain all data')


        returned_audio_data = queue_source1.get_audio_data(len(self.mock_audio_data))
        self.assertAlmostEqual(returned_audio_data.duration, 4.0)
        self.assertEqual(returned_audio_data.length, len(self.mock_audio_data)-11025)
        self.assertEqual(_bytes_from_audiodata(returned_audio_data), self.mock_audio_data[11025:], 'Should have seeked past 1 second')

    def test_seek_aligned_to_sample_size_2_bytes(self):
        self.create_valid_mock_source(bitrate=16, channels=1)
        static_source = StaticSource(self.mock_source)

        queue_source = static_source.get_queue_source()
        queue_source.seek(0.01)
        returned_audio_data = queue_source.get_audio_data(len(self.mock_audio_data))
        self.assertEqual(returned_audio_data.length % 2, 0, 'Must seek and return aligned to 2 byte chunks')

    def test_seek_aligned_to_sample_size_4_bytes(self):
        self.create_valid_mock_source(bitrate=16, channels=2)
        static_source = StaticSource(self.mock_source)

        queue_source = static_source.get_queue_source()
        queue_source.seek(0.01)
        returned_audio_data = queue_source.get_audio_data(len(self.mock_audio_data))
        self.assertEqual(returned_audio_data.length % 4, 0, 'Must seek and return aligned to 4 byte chunks')

    def test_empty_source(self):
        """Test that wrapping an empty source does not cause trouble."""
        self.mock_source = mock.MagicMock()
        self.mock_queue_source = self.mock_source.get_queue_source.return_value

        self.mock_queue_source.audio_format = None
        self.mock_queue_source.video_format = None
        self.mock_source.audio_format = None
        self.mock_source.video_format = None

        static_source = StaticSource(self.mock_source)

        self.assertEqual(static_source.duration, 0.)
        self.assertIsNone(static_source.get_queue_source())

    def test_not_directly_queueable(self):
        """A StaticSource cannot be played directly. Its queue source returns a playable object."""
        self.create_valid_mock_source(bitrate=16, channels=2)
        static_source = StaticSource(self.mock_source)

        with self.assertRaises(RuntimeError):
            static_source.get_audio_data(1024)

    def test_run_empty(self):
        """When running out of data, return None"""
        self.create_valid_mock_source()
        static_source = StaticSource(self.mock_source)
        queue_source = static_source.get_queue_source()
        returned_audio_data = queue_source.get_audio_data(self.mock_data_length)
        self.assertIsNotNone(returned_audio_data)

        no_more_audio_data = queue_source.get_audio_data(1024)
        self.assertIsNone(no_more_audio_data)


@unittest.skip('Too dangerous to modify 2.0.x\'s SourceGroups, this test will fail for them')
class SourceGroupTestCase(unittest.TestCase):
    def test_empty(self):
        group = SourceGroup()
        self.assertIsNone(group.get_audio_data(2048))

    def test_functionality(self):
        fake_data = ((b'a', 1000, 0.5), (b'b', 40000, 2.0), (b'c', 20000, 4.0), (b'd', 9992, 4.0))
        audio_data = [AudioData(b * l, l) for b, l, _ in fake_data]
        audio_data[0].timestamp = 1.23

        expected_data = b''.join(d * l for d, l, _ in fake_data)
        expected_duration = sum(d for _, _, d in fake_data)
        total_length = len(expected_data)

        sources = [mock.MagicMock(audio_format=AudioFormat(2, 8, 11025)) for _ in range(4)]
        exhausted = [False] * 4

        for i, mock_source in enumerate(sources):
            def _get_audio_data(_, j=i):
                if exhausted[j]:
                    return None
                exhausted[j] = True
                return audio_data[j]

            mock_source.duration = fake_data[i][2]
            mock_source.get_audio_data.side_effect = _get_audio_data
            mock_source.get_queue_source.return_value = mock_source

        group = SourceGroup()
        for mock_source in sources:
            group.add(mock_source)

        ret_data = group.get_audio_data(total_length)
        self.assertEqual(expected_data, ret_data.data)
        self.assertAlmostEqual(ret_data.timestamp, 1.23)
        self.assertAlmostEqual(ret_data.duration, expected_duration)

    def test_inequal_audio_format(self):
        source_a = mock.Mock(audio_format=AudioFormat(1, 8, 44100), duration=None)
        source_b = mock.Mock(audio_format=AudioFormat(2, 16, 44100), duration=None)
        source_a.get_queue_source.return_value = source_a
        source_b.get_queue_source.return_value = source_b

        group = SourceGroup()
        group.add(source_a)
        self.assertEqual(group.audio_format, source_a.audio_format)
        with self.assertRaises(MediaException):
            group.add(source_b)



#!/usr/bin/env python

'''Parse a C source file.

To use, subclass CParser and override its handle_* methods.  Then instantiate
the class with a string to parse.

Derived from ANSI C grammar:
  * Lexicon: http://www.lysator.liu.se/c/ANSI-C-grammar-l.html
  * Grammar: http://www.lysator.liu.se/c/ANSI-C-grammar-y.html

Reference is C99:
  * http://www.open-std.org/JTC1/SC22/WG14/www/docs/n1124.pdf

'''
import cPickle
import operator
import os.path
import re
import sys
import time
import warnings

import preprocessor
import yacc

tokens = (

    'PP_IF', 'PP_IFDEF', 'PP_IFNDEF', 'PP_ELIF', 'PP_ELSE',
    'PP_ENDIF', 'PP_INCLUDE', 'PP_DEFINE', 'PP_DEFINE_CONSTANT', 'PP_UNDEF',
    'PP_LINE', 'PP_ERROR', 'PP_PRAGMA',

    'IDENTIFIER', 'CONSTANT', 'CHARACTER_CONSTANT', 'STRING_LITERAL', 'SIZEOF',
    'PTR_OP', 'INC_OP', 'DEC_OP', 'LEFT_OP', 'RIGHT_OP', 'LE_OP', 'GE_OP',
    'EQ_OP', 'NE_OP', 'AND_OP', 'OR_OP', 'MUL_ASSIGN', 'DIV_ASSIGN',
    'MOD_ASSIGN', 'ADD_ASSIGN', 'SUB_ASSIGN', 'LEFT_ASSIGN', 'RIGHT_ASSIGN',
    'AND_ASSIGN', 'XOR_ASSIGN', 'OR_ASSIGN',  'HASH_HASH', 'PERIOD',
    'TYPE_NAME', 
    
    'TYPEDEF', 'EXTERN', 'STATIC', 'AUTO', 'REGISTER', 
    'CHAR', 'SHORT', 'INT', 'LONG', 'SIGNED', 'UNSIGNED', 'FLOAT', 'DOUBLE',
    'CONST', 'VOLATILE', 'VOID',
    'STRUCT', 'UNION', 'ENUM', 'ELLIPSIS',

    'CASE', 'DEFAULT', 'IF', 'ELSE', 'SWITCH', 'WHILE', 'DO', 'FOR', 'GOTO',
    'CONTINUE', 'BREAK', 'RETURN', '__ASM__'
)

keywords = [
    'auto', 'break', 'case', 'char', 'const', 'continue', 'default', 'do',
    'double', 'else', 'enum', 'extern', 'float', 'for', 'goto', 'if', 'int',
    'long', 'register', 'return', 'short', 'signed', 'sizeof', 'static',
    'struct', 'switch', 'typedef', 'union', 'unsigned', 'void', 'volatile',
    'while', '__asm__'
]


# --------------------------------------------------------------------------
# C Object Model
# --------------------------------------------------------------------------

class Declaration(object):
    def __init__(self):
        self.declarator = None
        self.type = Type()
        self.storage = None

    def __repr__(self):
        d = {
            'declarator': self.declarator,
            'type': self.type,
        }
        if self.storage:
            d['storage'] = self.storage
        l = ['%s=%r' % (k, v) for k, v in d.items()]
        return 'Declaration(%s)' % ', '.join(l)

class Declarator(object):
    pointer = None
    def __init__(self):
        self.identifier = None
        self.initializer = None
        self.array = None
        self.parameters = None

    # make pointer read-only to catch mistakes early
    pointer = property(lambda self: None)

    def __repr__(self):
        s = self.identifier or ''
        if self.array:
            s += repr(self.array)
        if self.initializer:
            s += ' = %r' % self.initializer
        if self.parameters is not None:
            s += '(' + ', '.join([repr(p) for p in self.parameters]) + ')'
        return s

class Pointer(Declarator):
    pointer = None
    def __init__(self):
        super().__init__()
        self.qualifiers = []

    def __repr__(self):
        q = ''
        if self.qualifiers:
            q = '<%s>' % ' '.join(self.qualifiers)
        return 'POINTER%s(%r)' % (q, self.pointer) + super().__repr__()

class Array(object):
    def __init__(self):
        self.size = None
        self.array = None

    def __repr__(self):
        if self.size:
            a =  '[%r]' % self.size
        else:
            a = '[]'
        if self.array:
            return repr(self.array) + a
        else:
            return a

class Parameter(object):
    def __init__(self):
        self.type = Type()
        self.storage = None
        self.declarator = None

    def __repr__(self):
        d = {
            'type': self.type,
        }
        if self.declarator:
            d['declarator'] = self.declarator
        if self.storage:
            d['storage'] = self.storage
        l = ['%s=%r' % (k, v) for k, v in d.items()]
        return 'Parameter(%s)' % ', '.join(l)


class Type(object):
    def __init__(self):
        self.qualifiers = []
        self.specifiers = []

    def __repr__(self):
        return ' '.join(self.qualifiers + [str(s) for s in self.specifiers])

# These are used only internally.

class StorageClassSpecifier(str):
    pass

class TypeSpecifier(str):
    pass

class StructTypeSpecifier(object):
    def __init__(self, is_union, tag, declarations):
        self.is_union = is_union
        self.tag = tag
        self.declarations = declarations

    def __repr__(self):
        if self.is_union:
            s = 'union'
        else:
            s = 'struct'
        if self.tag:
            s += ' %s' % self.tag
        if self.declarations:
            s += ' {%s}' % '; '.join([repr(d) for d in self.declarations])
        return s

class EnumSpecifier(object):
    def __init__(self, tag, enumerators):
        self.tag = tag
        self.enumerators = enumerators

    def __repr__(self):
        s = 'enum'
        if self.tag:
            s += ' %s' % self.tag
        if self.enumerators:
            s += ' {%s}' % ', '.join([repr(e) for e in self.enumerators])
        return s

class Enumerator(object):
    def __init__(self, name, expression):
        self.name = name
        self.expression = expression

    def __repr__(self):
        s = self.name
        if self.expression:
            s += ' = %r' % self.expression
        return s

class TypeQualifier(str):
    pass

def apply_specifiers(specifiers, declaration):
    '''Apply specifiers to the declaration (declaration may be
    a Parameter instead).'''
    for s in specifiers:
        if type(s) == StorageClassSpecifier:
            if declaration.storage:
                p.parser.cparser.handle_error(
                    'Declaration has more than one storage class', 
                    '???', p.lineno(1))
                return
            declaration.storage = s
        elif type(s) in (TypeSpecifier, StructTypeSpecifier, EnumSpecifier):
            declaration.type.specifiers.append(s)
        elif type(s) == TypeQualifier:
            declaration.type.qualifiers.append(s)


# --------------------------------------------------------------------------
# Expression Object Model
# --------------------------------------------------------------------------

class EvaluationContext(object):
    '''Interface for evaluating expression nodes.
    '''
    def evaluate_identifier(self, name):
        warnings.warn('Attempt to evaluate identifier "%s" failed' % name)
        return 0

    def evaluate_sizeof(self, type):
        warnings.warn('Attempt to evaluate sizeof "%s" failed' % str(type))
        return 0

class ExpressionNode(object):
    def evaluate(self, context):
        return 0

    def __str__(self):
        return ''

class ConstantExpressionNode(ExpressionNode):
    def __init__(self, value):
        self.value = value

    def evaluate(self, context):
        return self.value

    def __str__(self):
        return str(self.value)

class IdentifierExpressionNode(ExpressionNode):
    def __init__(self, name):
        self.name = name

    def evaluate(self, context):
        return context.evaluate_identifier(self.name)

    def __str__(self):
        return str(self.value)

class UnaryExpressionNode(ExpressionNode):
    def __init__(self, op, op_str, child):
        self.op = op
        self.op_str = op_str
        self.child = child

    def evaluate(self, context):
        return self.op(self.child.evaluate(context))

    def __str__(self):
        return '(%s %s)' % (self.op_str, self.child)

class SizeOfExpressionNode(ExpressionNode):
    def __init__(self, type):
        self.type = type

    def evaluate(self, context):
        return context.evaluate_sizeof(self.type)

    def __str__(self):
        return 'sizeof(%s)' % str(self.type)

class BinaryExpressionNode(ExpressionNode):
    def __init__(self, op, op_str, left, right):
        self.op = op
        self.op_str = op_str
        self.left = left
        self.right = right

    def evaluate(self, context):
        return self.op(self.left.evaluate(context), 
                       self.right.evaluate(context))

    def __str__(self):
        return '(%s %s %s)' % (self.left, self.op_str, self.right)

class LogicalAndExpressionNode(ExpressionNode):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def evaluate(self, context):
        return self.left.evaluate(context) and self.right.evaluate(context)

    def __str__(self):
        return '(%s && %s)' % (self.left, self.right)

class LogicalOrExpressionNode(ExpressionNode):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def evaluate(self, context):
        return self.left.evaluate(context) or self.right.evaluate(context)

    def __str__(self):
        return '(%s || %s)' % (self.left, self.right)

class ConditionalExpressionNode(ExpressionNode):
    def __init__(self, condition, left, right):
        self.condition = condition
        self.left = left
        self.right = right

    def evaluate(self, context):
        if self.condition.evaluate(context):
            return self.left.evaluate(context)
        else:
            return self.right.evaluate(context)

    def __str__(self):
        return '(%s ? %s : %s)' % (self.condition, self.left, self.right)


# --------------------------------------------------------------------------
# Grammar
# --------------------------------------------------------------------------

def p_translation_unit(p):
    '''translation_unit : 
                        | translation_unit external_declaration
    '''
    # Starting production.
    # Allow empty production so that files with no declarations are still
    #    valid.
    # Intentionally empty

def p_identifier(p):
    '''identifier : IDENTIFIER'''
    p[0] = IdentifierExpressionNode(p[1])

def p_constant(p):
    '''constant : CONSTANT
                | CHARACTER_CONSTANT
    '''
    def _to_int(s):
        s = s.rstrip('lLuU')
        if s.startswith('0x'):
            return int(s, base=16)
        elif s.startswith('0'):
            return int(s, base=8)
        else:
            return int(s)

    value = p[1]
    try:
        value = _to_int(value)
    except ValueError:
        pass
    p[0] = ConstantExpressionNode(value)

def p_string_literal(p):
    '''string_literal : STRING_LITERAL'''
    p[0] = ConstantExpressionNode(p[1])

def p_primary_expression(p):
    '''primary_expression : identifier
                          | constant
                          | string_literal
                          | '(' expression ')'
    '''
    if p[1] == '(':
        p[0] = p[2]
    else:
        p[0] = p[1]

def p_postfix_expression(p):
    '''postfix_expression : primary_expression
                  | postfix_expression '[' expression ']'
                  | postfix_expression '(' ')'
                  | postfix_expression '(' argument_expression_list ')'
                  | postfix_expression PERIOD IDENTIFIER
                  | postfix_expression PTR_OP IDENTIFIER
                  | postfix_expression INC_OP
                  | postfix_expression DEC_OP
    '''
    # XXX Largely unsupported
    p[0] = p[1]

def p_argument_expression_list(p):
    '''argument_expression_list : assignment_expression
                        | argument_expression_list ',' assignment_expression
    '''

def p_asm_expression(p):
    '''asm_expression : __ASM__ volatile_opt '(' string_literal ')'
                      | __ASM__ volatile_opt '(' string_literal ':' str_opt_expr_pair_list ')'
                      | __ASM__ volatile_opt '(' string_literal ':' str_opt_expr_pair_list ':' str_opt_expr_pair_list ')'
                      | __ASM__ volatile_opt '(' string_literal ':' str_opt_expr_pair_list ':' str_opt_expr_pair_list ':' str_opt_expr_pair_list ')'
    '''

    # Definitely not ISO C, adapted from example ANTLR GCC parser at
    #  http://www.antlr.org/grammar/cgram//grammars/GnuCParser.g
    # but more lenient (expressions permitted in optional final part, when
    # they shouldn't be -- avoids shift/reduce conflict with
    # str_opt_expr_pair_list).

    # XXX node not supported
    p[0] = ExpressionNode()

def p_str_opt_expr_pair_list(p):
    '''str_opt_expr_pair_list : 
                              | str_opt_expr_pair
                              | str_opt_expr_pair_list ',' str_opt_expr_pair
    '''

def p_str_opt_expr_pair(p):
   '''str_opt_expr_pair : string_literal
                        | string_literal '(' expression ')'
    '''

def p_volatile_opt(p):
    '''volatile_opt : 
                    | VOLATILE
    '''

def p_unary_expression(p):
    '''unary_expression : postfix_expression
                        | INC_OP unary_expression
                        | DEC_OP unary_expression
                        | unary_operator cast_expression
                        | SIZEOF unary_expression
                        | SIZEOF '(' type_name ')'
                        | asm_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    elif p[1] == 'sizeof':
        if p[2] == '(':
            p[0] = SizeOfExpressionNode(p[3])
        else:
            p[0] = SizeOfExpressionNode(p[2])
    elif type(p[1]) == tuple:
        # unary_operator reduces to (op, op_str)
        p[0] = UnaryExpressionNode(p[1][0], p[1][1], p[2])
    else:
        # XXX INC_OP and DEC_OP expression nodes not supported
        p[0] = p[2]

def p_unary_operator(p):
    '''unary_operator : '&'
                      | '*'
                      | '+'
                      | '-'
                      | '~'
                      | '!'
    '''
    # reduces to (op, op_str)
    p[0] = ({
        '+': operator.pos,
        '-': operator.neg,
        '~': operator.inv,
        '!': operator.not_,
        '&': 'AddressOfUnaryOperator',
        '*': 'DereferenceUnaryOperator'}[p[1]], p[1])

def p_cast_expression(p):
    '''cast_expression : unary_expression
                       | '(' type_name ')' cast_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        # XXX cast node not supported
        p[0] = p[4]

def p_multiplicative_expression(p):
    '''multiplicative_expression : cast_expression
                                 | multiplicative_expression '*' cast_expression
                                 | multiplicative_expression '/' cast_expression
                                 | multiplicative_expression '%' cast_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpressionNode({
            '*': operator.mul,
            '/': operator.div,
            '%': operator.mod}[p[2]], p[2], p[1], p[3])

def p_additive_expression(p):
    '''additive_expression : multiplicative_expression
                           | additive_expression '+' multiplicative_expression
                           | additive_expression '-' multiplicative_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpressionNode({
            '+': operator.add,
            '-': operator.sub}[p[2]], p[2], p[1], p[3])

def p_shift_expression(p):
    '''shift_expression : additive_expression
                        | shift_expression LEFT_OP additive_expression
                        | shift_expression RIGHT_OP additive_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpressionNode({
            '<<': operator.lshift,
            '>>': operator.rshift}[p[2]], p[2], p[1], p[3])
        
def p_relational_expression(p):
    '''relational_expression : shift_expression 
                             | relational_expression '<' shift_expression
                             | relational_expression '>' shift_expression
                             | relational_expression LE_OP shift_expression
                             | relational_expression GE_OP shift_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpressionNode({
            '>': operator.gt,
            '<': operator.lt,
            '<=': operator.le,
            '>=': operator.ge}[p[2]], p[2], p[1], p[3])

def p_equality_expression(p):
    '''equality_expression : relational_expression
                           | equality_expression EQ_OP relational_expression
                           | equality_expression NE_OP relational_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpressionNode({
            '==': operator.eq,
            '!=': operator.ne}[p[2]], p[2], p[1], p[3])

def p_and_expression(p):
    '''and_expression : equality_expression
                      | and_expression '&' equality_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpressionNode(operator.and_, '&', p[1], p[3])

def p_exclusive_or_expression(p):
    '''exclusive_or_expression : and_expression
                               | exclusive_or_expression '^' and_expression
    ''' 
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpressionNode(operator.xor, '^', p[1], p[3])

def p_inclusive_or_expression(p):
    '''inclusive_or_expression : exclusive_or_expression
                   | inclusive_or_expression '|' exclusive_or_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = BinaryExpressionNode(operator.or_, '|', p[1], p[3])

def p_logical_and_expression(p):
    '''logical_and_expression : inclusive_or_expression
                  | logical_and_expression AND_OP inclusive_or_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = LogicalAndExpressionNode(p[1], p[3])

def p_logical_or_expression(p):
    '''logical_or_expression : logical_and_expression
                  | logical_or_expression OR_OP logical_and_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = LogicalOrExpressionNode(p[1], p[3])


def p_conditional_expression(p):
    '''conditional_expression : logical_or_expression
          | logical_or_expression '?' expression ':' conditional_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = ConditionalExpressionNode(p[1], p[3], p[5])

def p_assignment_expression(p):
    '''assignment_expression : conditional_expression
                 | unary_expression assignment_operator assignment_expression
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        # XXX assignment expression node not supported
        p[0] = p[3]

def p_assignment_operator(p):
    '''assignment_operator : '='
                           | MUL_ASSIGN
                           | DIV_ASSIGN
                           | MOD_ASSIGN
                           | ADD_ASSIGN
                           | SUB_ASSIGN
                           | LEFT_ASSIGN
                           | RIGHT_ASSIGN
                           | AND_ASSIGN
                           | XOR_ASSIGN
                           | OR_ASSIGN
    '''

def p_expression(p):
    '''expression : assignment_expression
                  | expression ',' assignment_expression
    '''
    p[0] = p[1]
    # XXX sequence expression node not supported 

def p_constant_expression(p):
    '''constant_expression : conditional_expression
    '''
    p[0] = p[1]

def p_declaration(p):
    '''declaration : declaration_impl ';'
    '''
    # The ';' must be here, not in 'declaration', as declaration needs to
    # be executed before the ';' is shifted (otherwise the next lookahead will
    # be read, which may be affected by this declaration if its a typedef.

def p_declaration_impl(p):
    '''declaration_impl : declaration_specifiers
                        | declaration_specifiers init_declarator_list
    '''
    declaration = Declaration()
    apply_specifiers(p[1], declaration)

    if len(p) == 2:
        filename = p.slice[1].filename
        lineno = p.slice[1].lineno
        p.parser.cparser.impl_handle_declaration(declaration, filename, lineno)
        return

    filename = p.slice[2].filename
    lineno = p.slice[2].lineno
    for declarator in p[2]:
        declaration.declarator = declarator
        p.parser.cparser.impl_handle_declaration(declaration, filename, lineno)

""" # shift/reduce conflict with p_statement_error.
def p_declaration_error(p):
    '''declaration : error ';'
    '''
    # Error resynchronisation catch-all
"""

def p_declaration_specifiers(p):
    '''declaration_specifiers : storage_class_specifier
                              | storage_class_specifier declaration_specifiers
                              | type_specifier
                              | type_specifier declaration_specifiers
                              | type_qualifier
                              | type_qualifier declaration_specifiers
    '''
    if len(p) > 2:
        p[0] = (p[1],) + p[2]
    else:
        p[0] = (p[1],)

def p_init_declarator_list(p):
    '''init_declarator_list : init_declarator
                            | init_declarator_list ',' init_declarator
    '''
    if len(p) > 2:
        p[0] = p[1] + (p[3],)
    else:
        p[0] = (p[1],)

def p_init_declarator(p):
    '''init_declarator : declarator
                       | declarator '=' initializer
    '''
    p[0] = p[1]
    if len(p) > 2:
        p[0].initializer = p[2]

def p_storage_class_specifier(p):
    '''storage_class_specifier : TYPEDEF
                               | EXTERN
                               | STATIC
                               | AUTO
                               | REGISTER
    '''
    p[0] = StorageClassSpecifier(p[1])

def p_type_specifier(p):
    '''type_specifier : VOID
                      | CHAR
                      | SHORT
                      | INT
                      | LONG
                      | FLOAT
                      | DOUBLE
                      | SIGNED
                      | UNSIGNED
                      | struct_or_union_specifier
                      | enum_specifier
                      | TYPE_NAME
    '''
    if type(p[1]) in (StructTypeSpecifier, EnumSpecifier):
        p[0] = p[1]
    else:
        p[0] = TypeSpecifier(p[1])
    # TODO enum

def p_struct_or_union_specifier(p):
    '''struct_or_union_specifier : struct_or_union IDENTIFIER '{' struct_declaration_list '}'
         | struct_or_union TYPE_NAME '{' struct_declaration_list '}'
         | struct_or_union '{' struct_declaration_list '}'
         | struct_or_union IDENTIFIER
         | struct_or_union TYPE_NAME
    '''
    # The TYPE_NAME ones are dodgy, needed for Apple headers
    # CoreServices.framework/Frameworks/CarbonCore.framework/Headers/Files.h.
    # CoreServices.framework/Frameworks/OSServices.framework/Headers/Power.h
    if len(p) == 3:
        p[0] = StructTypeSpecifier(p[1], p[2], None)
    elif p[2] == '{':
        p[0] = StructTypeSpecifier(p[1], '', p[3])
    else:
        p[0] = StructTypeSpecifier(p[1], p[2], p[4])

def p_struct_or_union(p):
    '''struct_or_union : STRUCT
                       | UNION
    '''
    p[0] = p[1] == 'union'

def p_struct_declaration_list(p):
    '''struct_declaration_list : struct_declaration
                               | struct_declaration_list struct_declaration
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = p[1] + p[2]

def p_struct_declaration(p):
    '''struct_declaration : specifier_qualifier_list struct_declarator_list ';'
    '''
    # p[0] returned is a tuple, to handle multiple declarators in one
    # declaration.
    r = ()
    for declarator in p[2]:
        declaration = Declaration()
        apply_specifiers(p[1], declaration)
        declaration.declarator = declarator
        r += (declaration,)
    p[0] = r

def p_specifier_qualifier_list(p):
    '''specifier_qualifier_list : type_specifier specifier_qualifier_list
                                | type_specifier
                                | type_qualifier specifier_qualifier_list
                                | type_qualifier
    '''
    # XXX Interesting.. why is this one right-recursion?
    if len(p) == 3:
        p[0] = (p[1],) + p[2]
    else:
        p[0] = (p[1],)

def p_struct_declarator_list(p):
    '''struct_declarator_list : struct_declarator
                              | struct_declarator_list ',' struct_declarator
    '''
    if len(p) == 2:
        p[0] = (p[1],)
    else:
        p[0] = p[1] + (p[3],)

def p_struct_declarator(p):
    '''struct_declarator : declarator
                         | ':' constant_expression
                         | declarator ':' constant_expression
    '''
    # XXX ignoring bitfields.
    if p[1] == ':':
        p[0] = Declarator()
    else:
        p[0] = p[1]

def p_enum_specifier(p):
    '''enum_specifier : ENUM '{' enumerator_list '}'
                      | ENUM IDENTIFIER '{' enumerator_list '}'
                      | ENUM IDENTIFIER
    '''
    if len(p) == 5:
        p[0] = EnumSpecifier(None, p[3])
    elif len(p) == 6:
        p[0] = EnumSpecifier(p[2], p[4])
    else:
        p[0] = EnumSpecifier(p[2], ())

def p_enumerator_list(p):
    '''enumerator_list : enumerator_list_iso
                       | enumerator_list_iso ','
    '''
    # Apple headers sometimes have trailing ',' after enumerants, which is
    # not ISO C.
    p[0] = p[1]

def p_enumerator_list_iso(p):
    '''enumerator_list_iso : enumerator
                           | enumerator_list_iso ',' enumerator
    '''
    if len(p) == 2:
        p[0] = (p[1],)
    else:
        p[0] = p[1] + (p[3],)

def p_enumerator(p):
    '''enumerator : IDENTIFIER
                  | IDENTIFIER '=' constant_expression
    '''
    if len(p) == 2:
        p[0] = Enumerator(p[1], None)
    else:
        p[0] = Enumerator(p[1], p[3])

def p_type_qualifier(p):
    '''type_qualifier : CONST
                      | VOLATILE
    '''
    p[0] = TypeQualifier(p[1])

def p_declarator(p):
    '''declarator : pointer direct_declarator
                  | direct_declarator
    '''
    if len(p) > 2:
        p[0] = p[1]
        ptr = p[1]
        while ptr.pointer:
            ptr = ptr.pointer
        ptr.pointer = p[2]
    else:
        p[0] = p[1]

def p_direct_declarator(p):
    '''direct_declarator : IDENTIFIER
                         | TYPE_NAME
                         | '(' declarator ')'
                         | direct_declarator '[' constant_expression ']'
                         | direct_declarator '[' ']'
                         | direct_declarator '(' parameter_type_list ')'
                         | direct_declarator '(' identifier_list ')'
                         | direct_declarator '(' ')'
    '''
    # TYPE_NAME path is because some types are predefined in wrapper scripts;
    # it is not valid C.

    if isinstance(p[1], Declarator):
        p[0] = p[1] 
        if p[2] == '[':
            a = Array()
            a.array = p[0].array
            p[0].array = a
            if p[3] != ']':
                a.size = p[3]
        else:
            if p[3] == ')':
                p[0].parameters = ()
            else:
                p[0].parameters = p[3]
    elif p[1] == '(':
        p[0] = p[2]
    else:
        p[0] = Declarator()
        p[0].identifier = p[1]

    # Check parameters for (void) and simplify to empty tuple.
    if p[0].parameters and len(p[0].parameters) == 1:
        param = p[0].parameters[0]
        if param.type.specifiers == ['void'] and not param.declarator:
            p[0].parameters = ()


def p_pointer(p):
    '''pointer : '*'
               | '*' type_qualifier_list
               | '*' pointer
               | '*' type_qualifier_list pointer
    '''
    if len(p) == 2:
        p[0] = Pointer()
    elif len(p) == 3:
        if type(p[2]) == Pointer:
            p[0] = Pointer()
            p[0].pointer = p[2]
        else:
            p[0] = Pointer()
            p[0].qualifiers = p[2]
    else:
        p[0] = Pointer()
        p[0].qualifiers = p[2]
        p[0].pointer = p[3]

def p_type_qualifier_list(p):
    '''type_qualifier_list : type_qualifier
                           | type_qualifier_list type_qualifier
    '''
    if len(p) > 2:
        p[0] = p[1] + (p[2],)
    else:
        p[0] = (p[1],)

def p_parameter_type_list(p):
    '''parameter_type_list : parameter_list
                           | parameter_list ',' ELLIPSIS
    '''
    if len(p) > 2:
        p[0] = p[1] + (p[3],)
    else:
        p[0] = p[1]


def p_parameter_list(p):
    '''parameter_list : parameter_declaration
                      | parameter_list ',' parameter_declaration
    '''
    if len(p) > 2:
        p[0] = p[1] + (p[3],)
    else:
        p[0] = (p[1],)

def p_parameter_declaration(p):
    '''parameter_declaration : declaration_specifiers declarator
                             | declaration_specifiers abstract_declarator
                             | declaration_specifiers
    '''
    p[0] = Parameter()
    apply_specifiers(p[1], p[0])
    if len(p) > 2:
        p[0].declarator = p[2]

def p_identifier_list(p):
    '''identifier_list : IDENTIFIER
                       | identifier_list ',' IDENTIFIER
    '''
    param = Parameter()
    param.declarator = Declarator()
    if len(p) > 2:
        param.declarator.identifier = p[3]
        p[0] = p[1] + (param,)
    else:
        param.declarator.identifier = p[1]
        p[0] = (param,)

def p_type_name(p):
    '''type_name : specifier_qualifier_list
                 | specifier_qualifier_list abstract_declarator
    '''

def p_abstract_declarator(p):
    '''abstract_declarator : pointer
                           | direct_abstract_declarator
                           | pointer direct_abstract_declarator
    '''
    if len(p) == 2:
        p[0] = p[1]
        if type(p[0]) == Pointer:
            ptr = p[0]
            while ptr.pointer:
                ptr = ptr.pointer
            # Only if doesn't already terminate in a declarator
            if type(ptr) == Pointer:
                ptr.pointer = Declarator()
    else:
        p[0] = p[1]
        ptr = p[0]
        while ptr.pointer:
            ptr = ptr.pointer
        ptr.pointer = p[2]

def p_direct_abstract_declarator(p):
    '''direct_abstract_declarator : '(' abstract_declarator ')'
                      | '[' ']'
                      | '[' constant_expression ']'
                      | direct_abstract_declarator '[' ']'
                      | direct_abstract_declarator '[' constant_expression ']'
                      | '(' ')'
                      | '(' parameter_type_list ')'
                      | direct_abstract_declarator '(' ')'
                      | direct_abstract_declarator '(' parameter_type_list ')'
    '''
    if p[1] == '(' and isinstance(p[2], Declarator):
        p[0] = p[2]
    else:
        if isinstance(p[1], Declarator):
            p[0] = p[1]
            if p[2] == '[':
                a = Array()
                a.array = p[0].array
                p[0].array = a
                if p[3] != ']':
                    p[0].array.size = p[3]
            elif p[2] == '(':
                if p[3] == ')':
                    p[0].parameters = ()
                else:
                    p[0].parameters = p[3]
        else:
            p[0] = Declarator()
            if p[1] == '[':
                p[0].array = Array()
                if p[2] != ']':
                    p[0].array.size = p[2]
            elif p[1] == '(':
                if p[2] == ')':
                    p[0].parameters = ()
                else:
                    p[0].parameters = p[2]

def p_initializer(p):
    '''initializer : assignment_expression
                   | '{' initializer_list '}'
                   | '{' initializer_list ',' '}'
    '''

def p_initializer_list(p):
    '''initializer_list : initializer
                        | initializer_list ',' initializer
    '''

def p_statement(p):
    '''statement : labeled_statement
                 | compound_statement
                 | expression_statement
                 | selection_statement
                 | iteration_statement
                 | jump_statement
    '''

def p_labeled_statement(p):
    '''labeled_statement : IDENTIFIER ':' statement
                         | CASE constant_expression ':' statement
                         | DEFAULT ':' statement
    '''

def p_compound_statement(p):
    '''compound_statement : '{' '}'
                          | '{' statement_list '}'
                          | '{' declaration_list '}'
                          | '{' declaration_list statement_list '}'
    '''

def p_compound_statement_error(p):
    '''compound_statement : '{' error '}'
    '''
    # Error resynchronisation catch-all

def p_declaration_list(p):
    '''declaration_list : declaration
                        | declaration_list declaration
    '''

def p_statement_list(p):
    '''statement_list : statement
                      | statement_list statement
    '''

def p_expression_statement(p):
    '''expression_statement : ';'
                            | expression ';'
    '''
def p_expression_statement_error(p):
    '''expression_statement : error ';'
    '''
    # Error resynchronisation catch-all

def p_selection_statement(p):
    '''selection_statement : IF '(' expression ')' statement
                           | IF '(' expression ')' statement ELSE statement
                           | SWITCH '(' expression ')' statement
    '''

def p_iteration_statement(p):
    '''iteration_statement : WHILE '(' expression ')' statement
    | DO statement WHILE '(' expression ')' ';'
    | FOR '(' expression_statement expression_statement ')' statement
    | FOR '(' expression_statement expression_statement expression ')' statement
    '''	

def p_jump_statement(p):
    '''jump_statement : GOTO IDENTIFIER ';'
                      | CONTINUE ';'
                      | BREAK ';'
                      | RETURN ';'
                      | RETURN expression ';'
    '''

def p_external_declaration(p):
    '''external_declaration : declaration 
                            | function_definition
    '''

    # Intentionally empty

def p_function_definition(p):
    '''function_definition : declaration_specifiers declarator declaration_list compound_statement
                        | declaration_specifiers declarator compound_statement
                        | declarator declaration_list compound_statement
                        | declarator compound_statement
    '''

def p_error(t):
    if not t:
        # Crap, no way to get to CParser instance.  FIXME TODO
        print('Syntax error at end of file.', file=sys.stderr)
    else:
        t.lexer.cparser.handle_error('Syntax error at %r' % t.value, 
             t.filename, t.lineno)
    # Don't alter lexer: default behaviour is to pass error production
    # up until it hits the catch-all at declaration, at which point
    # parsing continues (synchronisation).

# --------------------------------------------------------------------------
# Lexer
# --------------------------------------------------------------------------

class CLexer(object):
    def __init__(self, cparser):
        self.cparser = cparser
        self.type_names = set()

    def input(self, tokens):
        self.tokens = tokens
        self.pos = 0

    def token(self):
        while self.pos < len(self.tokens):
            t = self.tokens[self.pos]
            self.pos += 1

            if not t:
                break

            # PP events
            if t.type == 'PP_DEFINE':
                name, value = t.value
                self.cparser.handle_define(
                    name, value, t.filename, t.lineno)
                continue
            elif t.type == 'PP_DEFINE_CONSTANT':
                name, value = t.value
                self.cparser.handle_define_constant(
                    name, value, t.filename, t.lineno)
                continue
            elif t.type == 'PP_IFNDEF':
                self.cparser.handle_ifndef(t.value, t.filename, t.lineno)
                continue
            # TODO: other PP tokens

            # Transform PP tokens into C tokens
            if t.type == 'LPAREN':
                t.type = '('
            elif t.type == 'PP_NUMBER':
                t.type = 'CONSTANT'
            elif t.type == 'IDENTIFIER' and t.value in keywords:
                t.type = t.value.upper()
            elif t.type == 'IDENTIFIER' and t.value in self.type_names:
                t.type = 'TYPE_NAME'
            t.lexer = self
            return t
        return None
        
# --------------------------------------------------------------------------
# Parser
# --------------------------------------------------------------------------

class CPreprocessorParser(preprocessor.PreprocessorParser):
    def __init__(self, cparser, **kwargs):
        self.cparser = cparser
        preprocessor.PreprocessorParser.__init__(self, **kwargs)

    def push_file(self, filename, data=None):
        if not self.cparser.handle_include(filename):
            return

        tokens = self.cparser.get_cached_tokens(filename)
        if tokens is not None:
            self.output += tokens
            return

        if not data:
            data = open(filename).read()
        self.lexer.push_input(data, filename)

class CParser(object):
    '''Parse a C source file.

    Subclass and override the handle_* methods.  Call `parse` with a string
    to parse.
    '''
    def __init__(self, stddef_types=True, gnu_types=True, cache_headers=True):
        self.preprocessor_parser = CPreprocessorParser(self)
        self.parser = yacc.Parser()
        yacc.yacc(method='LALR').init_parser(self.parser)
        self.parser.cparser = self

        self.lexer = CLexer(self)
        if stddef_types:
            self.lexer.type_names.add('wchar_t')
            self.lexer.type_names.add('ptrdiff_t')
            self.lexer.type_names.add('size_t')
        if gnu_types:
            self.lexer.type_names.add('__builtin_va_list')
        if sys.platform == 'win32':
            self.lexer.type_names.add('__int64')

        self.header_cache = {}
        self.cache_headers = cache_headers
        self.load_header_cache()
    
    def parse(self, filename, data=None, debug=False):
        '''Parse a file.  Give filename or filename + data.

        If `debug` is True, parsing state is dumped to stdout.
        '''
        if not data:
            data = open(filename, 'r').read()
        
        self.handle_status('Preprocessing %s' % filename)
        self.preprocessor_parser.parse(filename, data, debug=debug)
        self.lexer.input(self.preprocessor_parser.output)
        self.handle_status('Parsing %s' % filename)
        self.parser.parse(lexer=self.lexer, debug=debug)

    def load_header_cache(self, filename=None):
        if not filename:
            filename = '.header.cache'
        try:
            self.header_cache = cPickle.load(open(filename, 'rb'))
            self.handle_status('Loaded header cache "%s".  Found:' % filename)
            for header in self.header_cache.keys():
                self.handle_status('  %s' % header)
        except:
            self.handle_status('Failed to load header cache "%s"' % filename)

    def save_header_cache(self, filename=None):
        if not filename:
            filename = '.header.cache'
        try:
            cPickle.dump(self.header_cache, open(filename, 'wb'))
            self.handle_status('Updated header cache "%s"' % filename)
        except:
            self.handle_status('Failed to update header cache "%s"' % filename)

    def get_cached_tokens(self, header):
        '''Return a list of tokens for `header`.

        If there is no cached copy, return None.
        '''
        try:
            now = os.stat(header).st_mtime
        except OSError:
            now = time.time()
        current_memento = self.preprocessor_parser.get_memento()
        if header in self.header_cache:
            timestamp, memento, tokens, namespace = self.header_cache[header]
            if self.preprocessor_parser.system_headers:
                self.handle_status('Not using cached header "%s" because ' \
                                   'of overridden system_headers.' % header)
            elif now < timestamp:
                self.handle_status('Not using cached header "%s" because ' \
                                   'cached copy is stale.' % header)
            elif memento != current_memento:
                self.handle_status('Not using cached header "%s" because ' \
                                   'memento differs.' % header)
            else:
                self.handle_status('Using cached header "%s"' % header)
                self.preprocessor_parser.namespace = namespace
                return tokens

        if self.cache_headers and not self.preprocessor_parser.system_headers:
            self.handle_status('Caching header "%s"' % header)
            self.cache_headers = False
            ppp = preprocessor.PreprocessorParser()
            ppp.include_path = self.preprocessor_parser.include_path
            ppp.parse(filename=header,
                      namespace=self.preprocessor_parser.namespace)
            self.header_cache[header] = (now, current_memento, 
                                         ppp.output, ppp.namespace.copy())
            self.save_header_cache()
            self.cache_headers = True
            return ppp.output

        return None

    # ----------------------------------------------------------------------
    # Parser interface.  Override these methods in your subclass.
    # ----------------------------------------------------------------------

    def handle_error(self, message, filename, lineno):
        '''A parse error occurred.  
        
        The default implementation prints `lineno` and `message` to stderr.
        The parser will try to recover from errors by synchronising at the
        next semicolon.
        '''
        print('%s:%s %s' % (filename, lineno, message), file=sys.stderr)

    def handle_status(self, message):
        '''Progress information.

        The default implementationg prints message to stderr.
        '''
        print(message, file=sys.stderr)

    def handle_include(self, header):
        '''#include `header`
        
        Return True to proceed with including the header, otherwise return
        False to skip it.  The default implementation returns True.
        '''
        return True

    def handle_define(self, name, value, filename, lineno):
        '''#define `name` `value` 

        both are strings, value could not be parsed as an expression
        '''

    def handle_define_constant(self, name, value, filename, lineno):
        '''#define `name` `value`

        value is an int or float
        '''

    def handle_undef(self, name):
        '''#undef `name`'''

    def handle_if(self, expr):
        '''#if `expr`'''

    def handle_ifdef(self, name):
        '''#ifdef `name`'''

    def handle_ifndef(self, name, filename, lineno):
        '''#ifndef `name`'''

    def handle_elif(self, expr):
        '''#elif `expr`'''

    def handle_else(self):
        '''#else'''

    def handle_endif(self):
        '''#endif'''

    def impl_handle_declaration(self, declaration, filename, lineno):
        '''Internal method that calls `handle_declaration`.  This method
        also adds any new type definitions to the lexer's list of valid type
        names, which affects the parsing of subsequent declarations.
        '''
        if declaration.storage == 'typedef':
            declarator = declaration.declarator
            if not declarator:
                # XXX TEMPORARY while struct etc not filled
                return
            while declarator.pointer:
                declarator = declarator.pointer
            self.lexer.type_names.add(declarator.identifier)
        self.handle_declaration(declaration, filename, lineno)

    def handle_declaration(self, declaration, filename, lineno):
        '''A declaration was encountered.  
        
        `declaration` is an instance of Declaration.  Where a declaration has
        multiple initialisers, each is returned as a separate declaration.
        '''
        pass

class DebugCParser(CParser):
    '''A convenience class that prints each invocation of a handle_* method to
    stdout.
    '''
    def handle_include(self, header):
        print('#include header=%r' % header)
        return True

    def handle_define(self, name, value, filename, lineno):
        print('#define name=%r, value=%r' % (name, value))

    def handle_define_constant(self, name, value, filename, lineno):
        print('#define constant name=%r, value=%r' % (name, value))

    def handle_undef(self, name):
        print('#undef name=%r' % name)

    def handle_if(self, expr):
        print('#if expr=%s' % expr)

    def handle_ifdef(self, name):
        print('#ifdef name=%r' % name)

    def handle_ifndef(self, name, filename, lineno):
        print('#ifndef name=%r' % name)

    def handle_elif(self, expr):
        print('#elif expr=%s' % expr)

    def handle_else(self):
        print('#else')

    def handle_endif(self):
        print('#endif')

    def handle_declaration(self, declaration, filename, lineno):
        print(declaration)
        
if __name__ == '__main__':
    DebugCParser().parse(sys.argv[1], debug=True)



import sys

import reports
import fs
import mpexceptions


def description():
    template = ("""
Usage

    report.py sample report_name

Generates a report from the debugging info recorded while playing sample

Arguments

    sample: sample to report
    report_name: desired report, one of
{reports}

The report will be written to session's output dir under reports/sample.report_name.txt

Example

    report anomalies small.mp4

will write the report 'anomalies' to report/small.mp4.anomalies.txt
"""
                )
    text = template.format(reports=reports.available_reports_description(line_prefix=" " * 8))
    return text


def main(sample, report_name):
    try:
        pathserv = fs.get_path_info_for_active_session()
    except mpexceptions.ExceptionUndefinedSamplesDir:
        print("The env var 'pyglet_mp_samples_dir' is not defined.")
        return 1
    except mpexceptions.ExceptionNoSessionIsActive:
        print("*** Error, no session active.")
        return 1

    report_sample(pathserv, sample, report_name)


def report_sample(pathserv, sample, report_name):
    dbg_file = pathserv.dbg_filename(sample)
    outfile = pathserv.report_filename(sample, report_name)
    log_entries = fs.pickle_load(dbg_file)
    text = reports.report_by_name(report_name)(log_entries)
    fs.txt_save(text, outfile)


def sysargs_to_mainargs():
    """builds main args from sys.argv"""
    if len(sys.argv) < 3:
        print(description())
        sys.exit(1)
    report_name, sample = sys.argv[1:]
    return report_name, sample

if __name__ == '__main__':
    sample, report_name = sysargs_to_mainargs()
    main(sample, report_name)



"""Various utility functions used internally by pyglet
"""

import os
import math
import sys
from typing import Optional, Union, Callable

import pyglet
from pyglet.customtypes import Buffer


def asbytes(s: Union[str, Buffer]) -> bytes:
    if isinstance(s, bytes):
        return s
    elif isinstance(s, str):
        return bytes(ord(c) for c in s)
    else:
        return bytes(s)


def asstr(s: Optional[Union[str, Buffer]]) -> str:
    if s is None:
        return ''
    if isinstance(s, str):
        return s
    return s.decode("utf-8")  # type: ignore


# Keep these outside of the function since we don't need to re-define
# the function each time we make a call since no state is persisted.
def _debug_print_real(arg: str) -> bool:
    print(arg)
    return True


def _debug_print_dummy(arg: str) -> bool:
    return True


def debug_print(pyglet_option_name: str = 'debug') -> Callable[[str], bool]:
    """Get a debug printer controlled by the given ``pyglet.options`` name.

    This allows repurposing ``assert`` to write cleaner, more efficient
    debug output:

    #. Debug printers fit into a one-line ``assert`` statements instead
       of longer, slower key-lookup ``if`` statements
    #. Running Python with the ``-O`` flag makes pyglet run faster by
       skipping all ``assert`` statements

    Usage example::

        from pyglet.util import debug_print


        _debug_media = debug_print('debug_media')


        def some_func():
            # Python will skip the line below when run with -O
            assert _debug_media('My debug statement')

            # The rest of the function will run as normal
            ...

    For more information, please see `the Python command line
    documentation <https://docs.python.org/3/using/cmdline.html#cmdoption-O>`_.

    Args:
        `pyglet_option_name` :
            The name of a key in :attr:`pyglet.options` to read the
            debug flag's value from.

    Returns:
        A callable which prints a passed string and returns ``True``
        to allow auto-removal when running with ``-O``.

    """
    enabled = pyglet.options.get(pyglet_option_name, False)
    if enabled:
        return _debug_print_real
    return _debug_print_dummy


# Based on: https://stackoverflow.com/a/56225940
def closest_power_of_two(x: int) -> int:
    if x <= 2:
        return 2
    if (x >> (x.bit_length() - 2)) & 1:
        return 1 << math.ceil(math.log2(x))
    else:
        return 1 << math.floor(math.log2(x))


def next_or_equal_power_of_two(x: int) -> int:
    if x <= 1:
        return 1
    return 1 << math.ceil(math.log2(x))


class CodecRegistry:
    """Utility class for handling adding and querying of codecs."""

    def __init__(self):
        self._decoders = []
        self._encoders = []
        self._decoder_extensions = {}   # Map str -> list of matching Decoders
        self._encoder_extensions = {}   # Map str -> list of matching Encoders

    def get_encoders(self, filename=None):
        """Get a list of all encoders. If a `filename` is provided, only
        encoders supporting that extension will be returned. An empty list
        will be return if no encoders for that extension are available.
        """
        if filename:
            root, ext = os.path.splitext(filename)
            extension = ext if ext else root        # If only ".ext" is provided
            return self._encoder_extensions.get(extension.lower(), [])
        return self._encoders

    def get_decoders(self, filename=None):
        """Get a list of all decoders. If a `filename` is provided, only
        decoders supporting that extension will be returned. An empty list
        will be return if no encoders for that extension are available.
        """
        if filename:
            root, ext = os.path.splitext(filename)
            extension = ext if ext else root        # If only ".ext" is provided
            return self._decoder_extensions.get(extension.lower(), [])
        return self._decoders

    def add_decoders(self, module):
        """Add a decoder module.  The module must define `get_decoders`.  Once
        added, the appropriate decoders defined in the codec will be returned by
        CodecRegistry.get_decoders.
        """
        for decoder in module.get_decoders():
            self._decoders.append(decoder)
            for extension in decoder.get_file_extensions():
                if extension not in self._decoder_extensions:
                    self._decoder_extensions[extension] = []
                self._decoder_extensions[extension].append(decoder)

    def add_encoders(self, module):
        """Add an encoder module.  The module must define `get_encoders`.  Once
        added, the appropriate encoders defined in the codec will be returned by
        CodecRegistry.get_encoders.
        """
        for encoder in module.get_encoders():
            self._encoders.append(encoder)
            for extension in encoder.get_file_extensions():
                if extension not in self._encoder_extensions:
                    self._encoder_extensions[extension] = []
                self._encoder_extensions[extension].append(encoder)

    def decode(self, filename, file, **kwargs):
        """Attempt to decode a file, using the available registered decoders.
        Any decoders that match the file extension will be tried first. If no
        decoders match the extension, all decoders will then be tried in order.
        """
        first_exception = None

        for decoder in self.get_decoders(filename):
            try:
                return decoder.decode(filename, file, **kwargs)
            except DecodeException as e:
                if not first_exception:
                    first_exception = e
                if file:
                    file.seek(0)

        for decoder in self.get_decoders():
            try:
                return decoder.decode(filename, file, **kwargs)
            except DecodeException:
                if file:
                    file.seek(0)

        if not first_exception:
            raise DecodeException(f"No decoders available for this file type: {filename}")
        raise first_exception

    def encode(self, media, filename, file=None, **kwargs):
        """Attempt to encode a pyglet object to a specified format. All registered
        encoders that advertise support for the specific file extension will be tried.
        If no encoders are available, an EncodeException will be raised.
        """

        first_exception = None
        for encoder in self.get_encoders(filename):

            try:
                return encoder.encode(media, filename, file, **kwargs)
            except EncodeException as e:
                first_exception = first_exception or e

        if not first_exception:
            raise EncodeException(f"No Encoders are available for this extension: '{filename}'")
        raise first_exception


class Decoder:
    def get_file_extensions(self):
        """Return a list or tuple of accepted file extensions, e.g. ['.wav', '.ogg']
        Lower-case only.
        """
        raise NotImplementedError()

    def decode(self, *args, **kwargs):
        """Read and decode the given file object and return an appropriate
        pyglet object. Throws DecodeException if there is an error.
        `filename` can be a file type hint.
        """
        raise NotImplementedError()

    def __hash__(self):
        return hash(self.__class__.__name__)

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__

    def __repr__(self):
        return "{0}{1}".format(self.__class__.__name__, self.get_file_extensions())


class Encoder:
    def get_file_extensions(self):
        """Return a list or tuple of accepted file extensions, e.g. ['.wav', '.ogg']
        Lower-case only.
        """
        raise NotImplementedError()

    def encode(self, media, filename, file):
        """Encode the given media type to the given file.  `filename`
        provides a hint to the file format desired.  options are
        encoder-specific, and unknown options should be ignored or
        issue warnings.
        """
        raise NotImplementedError()

    def __hash__(self):
        return hash(self.__class__.__name__)

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__

    def __repr__(self):
        return "{0}{1}".format(self.__class__.__name__, self.get_file_extensions())


class DecodeException(Exception):
    __module__ = "CodecRegistry"


class EncodeException(Exception):
    __module__ = "CodecRegistry"
